"""
Some functions useful for the working with URLs and network.
"""

import requests
from typing import Optional
import re
import os
from mimetypes import guess_extension
from .string_tools import slugify


def is_url(url: str, allowed_url_prefixes=('http', 'ftp')) -> bool:
    """
    Check url for prefix match.
    """

    for prefix in set(allowed_url_prefixes):
        if url.startswith(prefix):
            return True

    return False


def download_from_url(url: str, timeout=None):
    """
    Download file from the URL.
    :param url: URL to download.
    :param timeout: timeout before fail.
    """

    try:
        response = requests.get(url, allow_redirects=True, timeout=timeout)
    except requests.exceptions.SSLError:
        print('Incorrect SSL certificate, trying to download without verifying...')
        response = requests.get(url, allow_redirects=True, verify=False,
                                timeout=timeout)

    if response.status_code != 200:
        raise OSError(str(response))

    return response


def get_filename_from_url(req: requests.Response) -> Optional[str]:
    """
    Get filename from url and, if not found, try to get from content-disposition.
    """

    if req.url.find('/'):
        result = req.url.rsplit('/', 1)[1]
    else:
        cd = req.headers.get('content-disposition')

        if cd is None:
            return None

        file_name = re.findall('filename=(.+)', cd)

        if len(file_name) == 0:
            return None

        result = file_name[0]

    f_name, f_ext = os.path.splitext(result)

    result = f'{slugify(f_name)}{guess_extension(req.headers["content-type"].partition(";")[0].strip())}' if not f_ext\
        else f'{slugify(f_name)}.{slugify(f_ext)}'

    return result



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.
# -*- coding: utf-8 -*-

from time import perf_counter
from typing import List, NamedTuple, Optional, Union
import time
from collections import namedtuple
from abc import ABC
import tabulate


class Timer(object):
    """A simple timer."""

    def __init__(self, name: str):
        self.name = name
        self.total_time = 0.
        self.calls = 0
        self.start_time = 0.
        self.diff = 0.
        self.average_time = 0.

    def tic(self):
        # using time.time instead of time.clock because time time.clock
        # does not normalize for multithreading
        self.start_time = time.perf_counter()

    def toc(self, average=True):
        self.diff = time.perf_counter() - self.start_time
        self.total_time += self.diff
        self.calls += 1
        self.average_time = self.total_time / self.calls
        if average:
            return self.average_time
        else:
            return self.diff

    def clear(self):
        self.total_time = 0.
        self.calls = 0
        self.start_time = 0.
        self.diff = 0.
        self.average_time = 0.


class TimerManager(ABC):

    def __init__(self, timer_names: Union[List[str], str]):
        self.timer_names = timer_names
        self.tpls = []

        if isinstance(self.timer_names, list):
            for tn in self.timer_names:
                setattr(self, tn, Timer(tn))
                self.tpls.append(tn)

    def collect_avg(self):
        for tn in self.timer_names:
            obj = getattr(self, tn)
            print(f'{obj.name}: {obj.average_time*1000:.3f}ms')
        print('')


'''

timers = create_some_timers(['det', 'mesh', 'reg'])

timers.det.tic()
...
timers.det.toc()


timers.mesh.tic()
...
timers.mesh.toc()


timers.collect_avg()

'''


class ATimer:
    records = {}
    tmp = None

    @classmethod
    def tic(cls):
        cls.tmp = time.time()

    @classmethod
    def toc(cls):
        res = (time.time() - cls.tmp) * 1000
        cls.tmp = None
        return res

    @classmethod
    def report(cls):
        header = ['', 'Time(ms)']
        contents = []
        for key, val in cls.records.items():
            contents.append(
                ['{:20s}'.format(key), '{:.2f}'.format(sum(val)/len(val))])
        print(tabulate.tabulate(contents, header, tablefmt='fancy_grid'))

    def __init__(self, name, silent=False):
        self.name = name
        self.silent = silent
        if name not in ATimer.records.keys():
            ATimer.records[name] = []

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, exc_type, exc_value, exc_tb):
        end = time.time()
        ATimer.records[self.name].append((end-self.start)*1000)
        if not self.silent:
            t = (end - self.start)*1000
            if t > 1000:
                print('-> [{:20s}]: {:5.1f}s'.format(self.name, t/1000))
            elif t > 1e3*60*60:
                print('-> [{:20s}]: {:5.1f}min'.format(self.name, t/1e3/60))
            else:
                print('-> [{:20s}]: {:5.1f}ms'.format(self.name,
                      (end-self.start)*1000))



#
# Copyright (c) 2021, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import tensorrt as trt
import numpy as np
import ctypes

try:
    import pycuda.driver as cuda
except ImportError as e:
    print('pycuda not installed, calibrator will be disabled.')



class Calibrator(trt.IInt8EntropyCalibrator2):
    '''calibrator
        IInt8EntropyCalibrator2
        IInt8LegacyCalibrator
        IInt8EntropyCalibrator
        IInt8MinMaxCalibrator

    '''
    def __init__(self, stream, cache_file=""):
        trt.IInt8EntropyCalibrator2.__init__(self)       
        self.stream = stream
        self.d_input = cuda.mem_alloc(self.stream.calibration_data.nbytes)
        self.cache_file = cache_file
        # print(self.cache_file)
        stream.reset()
        

    def get_batch_size(self):
        return self.stream.batch_size

    def get_batch(self, names):

        batch = self.stream.next_batch()
        if not batch.size:  
            return None

        cuda.memcpy_htod(self.d_input, batch)
        return [int(self.d_input)]

    def read_calibration_cache(self):
        # If there is a cache, use it instead of calibrating again. Otherwise, implicitly return None.
        if os.path.exists(self.cache_file):
            with open(self.cache_file, "rb") as f:
                print(f"[INFO] Using calibration cache to save time: {self.cache_file}")
                return f.read()

    def write_calibration_cache(self, cache): 
        with open(self.cache_file, "wb") as f:
            print(f"[INFO] Caching calibration data for future use: {self.cache_file}")
            f.write(cache)



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
This file contains some useful common label map
such as COCO, VOC, Cityscapes ...

this ids are same with official one.

"""

# NOTE: coco has some skip category, such as skipped 83...
# so we have to using map

coco_label_map = {
    0: '__background__',
    1: 'person',
    2: 'bicycle',
    3: 'car',
    4: 'motorcycle',
    5: 'airplane',
    6: 'bus',
    7: 'train',
    8: 'truck',
    9: 'boat',
    10: 'traffic light',
    11: 'fire hydrant',
    13: 'stop sign',
    14: 'parking meter',
    15: 'bench',
    16: 'bird',
    17: 'cat',
    18: 'dog',
    19: 'horse',
    20: 'sheep',
    21: 'cow',
    22: 'elephant',
    23: 'bear',
    24: 'zebra',
    25: 'giraffe',
    27: 'backpack',
    28: 'umbrella',
    31: 'handbag',
    32: 'tie',
    33: 'suitcase',
    34: 'frisbee',
    35: 'skis',
    36: 'snowboard',
    37: 'sports ball',
    38: 'kite',
    39: 'baseball bat',
    40: 'baseball glove',
    41: 'skateboard',
    42: 'surfboard',
    43: 'tennis racket',
    44: 'bottle',
    46: 'wine glass',
    47: 'cup',
    48: 'fork',
    49: 'knife',
    50: 'spoon',
    51: 'bowl',
    52: 'banana',
    53: 'apple',
    54: 'sandwich',
    55: 'orange',
    56: 'broccoli',
    57: 'carrot',
    58: 'hot dog',
    59: 'pizza',
    60: 'donut',
    61: 'cake',
    62: 'chair',
    63: 'couch',
    64: 'potted plant',
    65: 'bed',
    67: 'dining table',
    70: 'toilet',
    72: 'tv',
    73: 'laptop',
    74: 'mouse',
    75: 'remote',
    76: 'keyboard',
    77: 'cell phone',
    78: 'microwave',
    79: 'oven',
    80: 'toaster',
    81: 'sink',
    82: 'refrigerator',
    84: 'book',
    85: 'clock',
    86: 'vase',
    87: 'scissors',
    88: 'teddy bear',
    89: 'hair drier',
    90: 'toothbrush'
}
coco_stuff_list = ['things', 'banner', 'blanket', 'bridge', 'cardboard', 'counter', 'curtain', 'door-stuff', 'floor-wood', 'flower', 'fruit', 'gravel', 'house', 'light', 'mirror-stuff', 'net', 'pillow', 'platform', 'playingfield', 'railroad', 'river', 'road', 'roof', 'sand', 'sea',
                   'shelf', 'snow', 'stairs', 'tent', 'towel', 'wall-brick', 'wall-stone', 'wall-tile', 'wall-wood', 'water', 'window-blind', 'window', 'tree', 'fence', 'ceiling', 'sky', 'cabinet', 'table', 'floor', 'pavement', 'mountain', 'grass', 'dirt', 'paper', 'food', 'building', 'rock', 'wall', 'rug']
coco_label_map_list = list(coco_label_map.values())
coco_label_map_list.append(coco_stuff_list)

# map with mostly model prediction which is 0-80
coco_label_map_continuous = {
    i: coco_label_map_list[i] for i in range(0, len(coco_label_map_list))}


voc_label_map = {
    0: '_background_',
    1: 'aeroplane',
    2: 'bicycle',
    3: 'bird',
    4: 'boat',
    5: 'bottle',
    6: 'bus',
    7: 'car',
    8: 'cat',
    9: 'chair',
    10: 'cow',
    11: 'diningtable',
    12: 'dog',
    13: 'horse',
    14: 'motorbike',
    15: 'person',
    16: 'pottedplant',
    17: 'sheep',
    18: 'sofa',
    19: 'train',
    20: 'tvmonitor',
}

voc_label_map_list = list(voc_label_map.values())

cityscapes_label_map = {
    0: 'unlabeled',
    1: 'ego vehicle',
    2: 'rectification border',
    3: 'out of roi',
    4: 'static',
    5: 'dynamic',
    6: 'ground',
    7: 'road',
    8: 'sidewalk',
    9: 'parking',
    10: 'rail track',
    11: 'building',
    12: 'wall',
    13: 'fence',
    14: 'guard rail',
    15: 'bridge',
    16: 'tunnel',
    17: 'pole',
    18: 'polegroup',
    19: 'traffic light',
    20: 'traffic sign',
    21: 'vegetation',
    22: 'terrain',
    23: 'sky',
    24: 'person',
    25: 'rider',
    26: 'car',
    27: 'truck',
    28: 'bus',
    29: 'caravan',
    30: 'trailer',
    31: 'train',
    32: 'motorcycle',
    33: 'bicycle',
    -1: 'license plate',
}

imagenet_label_map = {0: 'tench, Tinca tinca',
                      1: 'goldfish, Carassius auratus',
                      2: 'great white shark, white shark, man-eater, man-eating shark, Carcharodon carcharias',
                      3: 'tiger shark, Galeocerdo cuvieri',
                      4: 'hammerhead, hammerhead shark',
                      5: 'electric ray, crampfish, numbfish, torpedo',
                      6: 'stingray',
                      7: 'cock',
                      8: 'hen',
                      9: 'ostrich, Struthio camelus',
                      10: 'brambling, Fringilla montifringilla',
                      11: 'goldfinch, Carduelis carduelis',
                      12: 'house finch, linnet, Carpodacus mexicanus',
                      13: 'junco, snowbird',
                      14: 'indigo bunting, indigo finch, indigo bird, Passerina cyanea',
                      15: 'robin, American robin, Turdus migratorius',
                      16: 'bulbul',
                      17: 'jay',
                      18: 'magpie',
                      19: 'chickadee',
                      20: 'water ouzel, dipper',
                      21: 'kite',
                      22: 'bald eagle, American eagle, Haliaeetus leucocephalus',
                      23: 'vulture',
                      24: 'great grey owl, great gray owl, Strix nebulosa',
                      25: 'European fire salamander, Salamandra salamandra',
                      26: 'common newt, Triturus vulgaris',
                      27: 'eft',
                      28: 'spotted salamander, Ambystoma maculatum',
                      29: 'axolotl, mud puppy, Ambystoma mexicanum',
                      30: 'bullfrog, Rana catesbeiana',
                      31: 'tree frog, tree-frog',
                      32: 'tailed frog, bell toad, ribbed toad, tailed toad, Ascaphus trui',
                      33: 'loggerhead, loggerhead turtle, Caretta caretta',
                      34: 'leatherback turtle, leatherback, leathery turtle, Dermochelys coriacea',
                      35: 'mud turtle',
                      36: 'terrapin',
                      37: 'box turtle, box tortoise',
                      38: 'banded gecko',
                      39: 'common iguana, iguana, Iguana iguana',
                      40: 'American chameleon, anole, Anolis carolinensis',
                      41: 'whiptail, whiptail lizard',
                      42: 'agama',
                      43: 'frilled lizard, Chlamydosaurus kingi',
                      44: 'alligator lizard',
                      45: 'Gila monster, Heloderma suspectum',
                      46: 'green lizard, Lacerta viridis',
                      47: 'African chameleon, Chamaeleo chamaeleon',
                      48: 'Komodo dragon, Komodo lizard, dragon lizard, giant lizard, Varanus komodoensis',
                      49: 'African crocodile, Nile crocodile, Crocodylus niloticus',
                      50: 'American alligator, Alligator mississipiensis',
                      51: 'triceratops',
                      52: 'thunder snake, worm snake, Carphophis amoenus',
                      53: 'ringneck snake, ring-necked snake, ring snake',
                      54: 'hognose snake, puff adder, sand viper',
                      55: 'green snake, grass snake',
                      56: 'king snake, kingsnake',
                      57: 'garter snake, grass snake',
                      58: 'water snake',
                      59: 'vine snake',
                      60: 'night snake, Hypsiglena torquata',
                      61: 'boa constrictor, Constrictor constrictor',
                      62: 'rock python, rock snake, Python sebae',
                      63: 'Indian cobra, Naja naja',
                      64: 'green mamba',
                      65: 'sea snake',
                      66: 'horned viper, cerastes, sand viper, horned asp, Cerastes cornutus',
                      67: 'diamondback, diamondback rattlesnake, Crotalus adamanteus',
                      68: 'sidewinder, horned rattlesnake, Crotalus cerastes',
                      69: 'trilobite',
                      70: 'harvestman, daddy longlegs, Phalangium opilio',
                      71: 'scorpion',
                      72: 'black and gold garden spider, Argiope aurantia',
                      73: 'barn spider, Araneus cavaticus',
                      74: 'garden spider, Aranea diademata',
                      75: 'black widow, Latrodectus mactans',
                      76: 'tarantula',
                      77: 'wolf spider, hunting spider',
                      78: 'tick',
                      79: 'centipede',
                      80: 'black grouse',
                      81: 'ptarmigan',
                      82: 'ruffed grouse, partridge, Bonasa umbellus',
                      83: 'prairie chicken, prairie grouse, prairie fowl',
                      84: 'peacock',
                      85: 'quail',
                      86: 'partridge',
                      87: 'African grey, African gray, Psittacus erithacus',
                      88: 'macaw',
                      89: 'sulphur-crested cockatoo, Kakatoe galerita, Cacatua galerita',
                      90: 'lorikeet',
                      91: 'coucal',
                      92: 'bee eater',
                      93: 'hornbill',
                      94: 'hummingbird',
                      95: 'jacamar',
                      96: 'toucan',
                      97: 'drake',
                      98: 'red-breasted merganser, Mergus serrator',
                      99: 'goose',
                      100: 'black swan, Cygnus atratus',
                      101: 'tusker',
                      102: 'echidna, spiny anteater, anteater',
                      103: 'platypus, duckbill, duckbilled platypus, duck-billed platypus, Ornithorhynchus anatinus',
                      104: 'wallaby, brush kangaroo',
                      105: 'koala, koala bear, kangaroo bear, native bear, Phascolarctos cinereus',
                      106: 'wombat',
                      107: 'jellyfish',
                      108: 'sea anemone, anemone',
                      109: 'brain coral',
                      110: 'flatworm, platyhelminth',
                      111: 'nematode, nematode worm, roundworm',
                      112: 'conch',
                      113: 'snail',
                      114: 'slug',
                      115: 'sea slug, nudibranch',
                      116: 'chiton, coat-of-mail shell, sea cradle, polyplacophore',
                      117: 'chambered nautilus, pearly nautilus, nautilus',
                      118: 'Dungeness crab, Cancer magister',
                      119: 'rock crab, Cancer irroratus',
                      120: 'fiddler crab',
                      121: 'king crab, Alaska crab, Alaskan king crab, Alaska king crab, Paralithodes camtschatica',
                      122: 'American lobster, Northern lobster, Maine lobster, Homarus americanus',
                      123: 'spiny lobster, langouste, rock lobster, crawfish, crayfish, sea crawfish',
                      124: 'crayfish, crawfish, crawdad, crawdaddy',
                      125: 'hermit crab',
                      126: 'isopod',
                      127: 'white stork, Ciconia ciconia',
                      128: 'black stork, Ciconia nigra',
                      129: 'spoonbill',
                      130: 'flamingo',
                      131: 'little blue heron, Egretta caerulea',
                      132: 'American egret, great white heron, Egretta albus',
                      133: 'bittern',
                      134: 'crane',
                      135: 'limpkin, Aramus pictus',
                      136: 'European gallinule, Porphyrio porphyrio',
                      137: 'American coot, marsh hen, mud hen, water hen, Fulica americana',
                      138: 'bustard',
                      139: 'ruddy turnstone, Arenaria interpres',
                      140: 'red-backed sandpiper, dunlin, Erolia alpina',
                      141: 'redshank, Tringa totanus',
                      142: 'dowitcher',
                      143: 'oystercatcher, oyster catcher',
                      144: 'pelican',
                      145: 'king penguin, Aptenodytes patagonica',
                      146: 'albatross, mollymawk',
                      147: 'grey whale, gray whale, devilfish, Eschrichtius gibbosus, Eschrichtius robustus',
                      148: 'killer whale, killer, orca, grampus, sea wolf, Orcinus orca',
                      149: 'dugong, Dugong dugon',
                      150: 'sea lion',
                      151: 'Chihuahua',
                      152: 'Japanese spaniel',
                      153: 'Maltese dog, Maltese terrier, Maltese',
                      154: 'Pekinese, Pekingese, Peke',
                      155: 'Shih-Tzu',
                      156: 'Blenheim spaniel',
                      157: 'papillon',
                      158: 'toy terrier',
                      159: 'Rhodesian ridgeback',
                      160: 'Afghan hound, Afghan',
                      161: 'basset, basset hound',
                      162: 'beagle',
                      163: 'bloodhound, sleuthhound',
                      164: 'bluetick',
                      165: 'black-and-tan coonhound',
                      166: 'Walker hound, Walker foxhound',
                      167: 'English foxhound',
                      168: 'redbone',
                      169: 'borzoi, Russian wolfhound',
                      170: 'Irish wolfhound',
                      171: 'Italian greyhound',
                      172: 'whippet',
                      173: 'Ibizan hound, Ibizan Podenco',
                      174: 'Norwegian elkhound, elkhound',
                      175: 'otterhound, otter hound',
                      176: 'Saluki, gazelle hound',
                      177: 'Scottish deerhound, deerhound',
                      178: 'Weimaraner',
                      179: 'Staffordshire bullterrier, Staffordshire bull terrier',
                      180: 'American Staffordshire terrier, Staffordshire terrier, American pit bull terrier, pit bull terrier',
                      181: 'Bedlington terrier',
                      182: 'Border terrier',
                      183: 'Kerry blue terrier',
                      184: 'Irish terrier',
                      185: 'Norfolk terrier',
                      186: 'Norwich terrier',
                      187: 'Yorkshire terrier',
                      188: 'wire-haired fox terrier',
                      189: 'Lakeland terrier',
                      190: 'Sealyham terrier, Sealyham',
                      191: 'Airedale, Airedale terrier',
                      192: 'cairn, cairn terrier',
                      193: 'Australian terrier',
                      194: 'Dandie Dinmont, Dandie Dinmont terrier',
                      195: 'Boston bull, Boston terrier',
                      196: 'miniature schnauzer',
                      197: 'giant schnauzer',
                      198: 'standard schnauzer',
                      199: 'Scotch terrier, Scottish terrier, Scottie',
                      200: 'Tibetan terrier, chrysanthemum dog',
                      201: 'silky terrier, Sydney silky',
                      202: 'soft-coated wheaten terrier',
                      203: 'West Highland white terrier',
                      204: 'Lhasa, Lhasa apso',
                      205: 'flat-coated retriever',
                      206: 'curly-coated retriever',
                      207: 'golden retriever',
                      208: 'Labrador retriever',
                      209: 'Chesapeake Bay retriever',
                      210: 'German short-haired pointer',
                      211: 'vizsla, Hungarian pointer',
                      212: 'English setter',
                      213: 'Irish setter, red setter',
                      214: 'Gordon setter',
                      215: 'Brittany spaniel',
                      216: 'clumber, clumber spaniel',
                      217: 'English springer, English springer spaniel',
                      218: 'Welsh springer spaniel',
                      219: 'cocker spaniel, English cocker spaniel, cocker',
                      220: 'Sussex spaniel',
                      221: 'Irish water spaniel',
                      222: 'kuvasz',
                      223: 'schipperke',
                      224: 'groenendael',
                      225: 'malinois',
                      226: 'briard',
                      227: 'kelpie',
                      228: 'komondor',
                      229: 'Old English sheepdog, bobtail',
                      230: 'Shetland sheepdog, Shetland sheep dog, Shetland',
                      231: 'collie',
                      232: 'Border collie',
                      233: 'Bouvier des Flandres, Bouviers des Flandres',
                      234: 'Rottweiler',
                      235: 'German shepherd, German shepherd dog, German police dog, alsatian',
                      236: 'Doberman, Doberman pinscher',
                      237: 'miniature pinscher',
                      238: 'Greater Swiss Mountain dog',
                      239: 'Bernese mountain dog',
                      240: 'Appenzeller',
                      241: 'EntleBucher',
                      242: 'boxer',
                      243: 'bull mastiff',
                      244: 'Tibetan mastiff',
                      245: 'French bulldog',
                      246: 'Great Dane',
                      247: 'Saint Bernard, St Bernard',
                      248: 'Eskimo dog, husky',
                      249: 'malamute, malemute, Alaskan malamute',
                      250: 'Siberian husky',
                      251: 'dalmatian, coach dog, carriage dog',
                      252: 'affenpinscher, monkey pinscher, monkey dog',
                      253: 'basenji',
                      254: 'pug, pug-dog',
                      255: 'Leonberg',
                      256: 'Newfoundland, Newfoundland dog',
                      257: 'Great Pyrenees',
                      258: 'Samoyed, Samoyede',
                      259: 'Pomeranian',
                      260: 'chow, chow chow',
                      261: 'keeshond',
                      262: 'Brabancon griffon',
                      263: 'Pembroke, Pembroke Welsh corgi',
                      264: 'Cardigan, Cardigan Welsh corgi',
                      265: 'toy poodle',
                      266: 'miniature poodle',
                      267: 'standard poodle',
                      268: 'Mexican hairless',
                      269: 'timber wolf, grey wolf, gray wolf, Canis lupus',
                      270: 'white wolf, Arctic wolf, Canis lupus tundrarum',
                      271: 'red wolf, maned wolf, Canis rufus, Canis niger',
                      272: 'coyote, prairie wolf, brush wolf, Canis latrans',
                      273: 'dingo, warrigal, warragal, Canis dingo',
                      274: 'dhole, Cuon alpinus',
                      275: 'African hunting dog, hyena dog, Cape hunting dog, Lycaon pictus',
                      276: 'hyena, hyaena',
                      277: 'red fox, Vulpes vulpes',
                      278: 'kit fox, Vulpes macrotis',
                      279: 'Arctic fox, white fox, Alopex lagopus',
                      280: 'grey fox, gray fox, Urocyon cinereoargenteus',
                      281: 'tabby, tabby cat',
                      282: 'tiger cat',
                      283: 'Persian cat',
                      284: 'Siamese cat, Siamese',
                      285: 'Egyptian cat',
                      286: 'cougar, puma, catamount, mountain lion, painter, panther, Felis concolor',
                      287: 'lynx, catamount',
                      288: 'leopard, Panthera pardus',
                      289: 'snow leopard, ounce, Panthera uncia',
                      290: 'jaguar, panther, Panthera onca, Felis onca',
                      291: 'lion, king of beasts, Panthera leo',
                      292: 'tiger, Panthera tigris',
                      293: 'cheetah, chetah, Acinonyx jubatus',
                      294: 'brown bear, bruin, Ursus arctos',
                      295: 'American black bear, black bear, Ursus americanus, Euarctos americanus',
                      296: 'ice bear, polar bear, Ursus Maritimus, Thalarctos maritimus',
                      297: 'sloth bear, Melursus ursinus, Ursus ursinus',
                      298: 'mongoose',
                      299: 'meerkat, mierkat',
                      300: 'tiger beetle',
                      301: 'ladybug, ladybeetle, lady beetle, ladybird, ladybird beetle',
                      302: 'ground beetle, carabid beetle',
                      303: 'long-horned beetle, longicorn, longicorn beetle',
                      304: 'leaf beetle, chrysomelid',
                      305: 'dung beetle',
                      306: 'rhinoceros beetle',
                      307: 'weevil',
                      308: 'fly',
                      309: 'bee',
                      310: 'ant, emmet, pismire',
                      311: 'grasshopper, hopper',
                      312: 'cricket',
                      313: 'walking stick, walkingstick, stick insect',
                      314: 'cockroach, roach',
                      315: 'mantis, mantid',
                      316: 'cicada, cicala',
                      317: 'leafhopper',
                      318: 'lacewing, lacewing fly',
                      319: "dragonfly, darning needle, devil's darning needle, sewing needle, snake feeder, snake doctor, mosquito hawk, skeeter hawk",
                      320: 'damselfly',
                      321: 'admiral',
                      322: 'ringlet, ringlet butterfly',
                      323: 'monarch, monarch butterfly, milkweed butterfly, Danaus plexippus',
                      324: 'cabbage butterfly',
                      325: 'sulphur butterfly, sulfur butterfly',
                      326: 'lycaenid, lycaenid butterfly',
                      327: 'starfish, sea star',
                      328: 'sea urchin',
                      329: 'sea cucumber, holothurian',
                      330: 'wood rabbit, cottontail, cottontail rabbit',
                      331: 'hare',
                      332: 'Angora, Angora rabbit',
                      333: 'hamster',
                      334: 'porcupine, hedgehog',
                      335: 'fox squirrel, eastern fox squirrel, Sciurus niger',
                      336: 'marmot',
                      337: 'beaver',
                      338: 'guinea pig, Cavia cobaya',
                      339: 'sorrel',
                      340: 'zebra',
                      341: 'hog, pig, grunter, squealer, Sus scrofa',
                      342: 'wild boar, boar, Sus scrofa',
                      343: 'warthog',
                      344: 'hippopotamus, hippo, river horse, Hippopotamus amphibius',
                      345: 'ox',
                      346: 'water buffalo, water ox, Asiatic buffalo, Bubalus bubalis',
                      347: 'bison',
                      348: 'ram, tup',
                      349: 'bighorn, bighorn sheep, cimarron, Rocky Mountain bighorn, Rocky Mountain sheep, Ovis canadensis',
                      350: 'ibex, Capra ibex',
                      351: 'hartebeest',
                      352: 'impala, Aepyceros melampus',
                      353: 'gazelle',
                      354: 'Arabian camel, dromedary, Camelus dromedarius',
                      355: 'llama',
                      356: 'weasel',
                      357: 'mink',
                      358: 'polecat, fitch, foulmart, foumart, Mustela putorius',
                      359: 'black-footed ferret, ferret, Mustela nigripes',
                      360: 'otter',
                      361: 'skunk, polecat, wood pussy',
                      362: 'badger',
                      363: 'armadillo',
                      364: 'three-toed sloth, ai, Bradypus tridactylus',
                      365: 'orangutan, orang, orangutang, Pongo pygmaeus',
                      366: 'gorilla, Gorilla gorilla',
                      367: 'chimpanzee, chimp, Pan troglodytes',
                      368: 'gibbon, Hylobates lar',
                      369: 'siamang, Hylobates syndactylus, Symphalangus syndactylus',
                      370: 'guenon, guenon monkey',
                      371: 'patas, hussar monkey, Erythrocebus patas',
                      372: 'baboon',
                      373: 'macaque',
                      374: 'langur',
                      375: 'colobus, colobus monkey',
                      376: 'proboscis monkey, Nasalis larvatus',
                      377: 'marmoset',
                      378: 'capuchin, ringtail, Cebus capucinus',
                      379: 'howler monkey, howler',
                      380: 'titi, titi monkey',
                      381: 'spider monkey, Ateles geoffroyi',
                      382: 'squirrel monkey, Saimiri sciureus',
                      383: 'Madagascar cat, ring-tailed lemur, Lemur catta',
                      384: 'indri, indris, Indri indri, Indri brevicaudatus',
                      385: 'Indian elephant, Elephas maximus',
                      386: 'African elephant, Loxodonta africana',
                      387: 'lesser panda, red panda, panda, bear cat, cat bear, Ailurus fulgens',
                      388: 'giant panda, panda, panda bear, coon bear, Ailuropoda melanoleuca',
                      389: 'barracouta, snoek',
                      390: 'eel',
                      391: 'coho, cohoe, coho salmon, blue jack, silver salmon, Oncorhynchus kisutch',
                      392: 'rock beauty, Holocanthus tricolor',
                      393: 'anemone fish',
                      394: 'sturgeon',
                      395: 'gar, garfish, garpike, billfish, Lepisosteus osseus',
                      396: 'lionfish',
                      397: 'puffer, pufferfish, blowfish, globefish',
                      398: 'abacus',
                      399: 'abaya',
                      400: "academic gown, academic robe, judge's robe",
                      401: 'accordion, piano accordion, squeeze box',
                      402: 'acoustic guitar',
                      403: 'aircraft carrier, carrier, flattop, attack aircraft carrier',
                      404: 'airliner',
                      405: 'airship, dirigible',
                      406: 'altar',
                      407: 'ambulance',
                      408: 'amphibian, amphibious vehicle',
                      409: 'analog clock',
                      410: 'apiary, bee house',
                      411: 'apron',
                      412: 'ashcan, trash can, garbage can, wastebin, ash bin, ash-bin, ashbin, dustbin, trash barrel, trash bin',
                      413: 'assault rifle, assault gun',
                      414: 'backpack, back pack, knapsack, packsack, rucksack, haversack',
                      415: 'bakery, bakeshop, bakehouse',
                      416: 'balance beam, beam',
                      417: 'balloon',
                      418: 'ballpoint, ballpoint pen, ballpen, Biro',
                      419: 'Band Aid',
                      420: 'banjo',
                      421: 'bannister, banister, balustrade, balusters, handrail',
                      422: 'barbell',
                      423: 'barber chair',
                      424: 'barbershop',
                      425: 'barn',
                      426: 'barometer',
                      427: 'barrel, cask',
                      428: 'barrow, garden cart, lawn cart, wheelbarrow',
                      429: 'baseball',
                      430: 'basketball',
                      431: 'bassinet',
                      432: 'bassoon',
                      433: 'bathing cap, swimming cap',
                      434: 'bath towel',
                      435: 'bathtub, bathing tub, bath, tub',
                      436: 'beach wagon, station wagon, wagon, estate car, beach waggon, station waggon, waggon',
                      437: 'beacon, lighthouse, beacon light, pharos',
                      438: 'beaker',
                      439: 'bearskin, busby, shako',
                      440: 'beer bottle',
                      441: 'beer glass',
                      442: 'bell cote, bell cot',
                      443: 'bib',
                      444: 'bicycle-built-for-two, tandem bicycle, tandem',
                      445: 'bikini, two-piece',
                      446: 'binder, ring-binder',
                      447: 'binoculars, field glasses, opera glasses',
                      448: 'birdhouse',
                      449: 'boathouse',
                      450: 'bobsled, bobsleigh, bob',
                      451: 'bolo tie, bolo, bola tie, bola',
                      452: 'bonnet, poke bonnet',
                      453: 'bookcase',
                      454: 'bookshop, bookstore, bookstall',
                      455: 'bottlecap',
                      456: 'bow',
                      457: 'bow tie, bow-tie, bowtie',
                      458: 'brass, memorial tablet, plaque',
                      459: 'brassiere, bra, bandeau',
                      460: 'breakwater, groin, groyne, mole, bulwark, seawall, jetty',
                      461: 'breastplate, aegis, egis',
                      462: 'broom',
                      463: 'bucket, pail',
                      464: 'buckle',
                      465: 'bulletproof vest',
                      466: 'bullet train, bullet',
                      467: 'butcher shop, meat market',
                      468: 'cab, hack, taxi, taxicab',
                      469: 'caldron, cauldron',
                      470: 'candle, taper, wax light',
                      471: 'cannon',
                      472: 'canoe',
                      473: 'can opener, tin opener',
                      474: 'cardigan',
                      475: 'car mirror',
                      476: 'carousel, carrousel, merry-go-round, roundabout, whirligig',
                      477: "carpenter's kit, tool kit",
                      478: 'carton',
                      479: 'car wheel',
                      480: 'cash machine, cash dispenser, automated teller machine, automatic teller machine, automated teller, automatic teller, ATM',
                      481: 'cassette',
                      482: 'cassette player',
                      483: 'castle',
                      484: 'catamaran',
                      485: 'CD player',
                      486: 'cello, violoncello',
                      487: 'cellular telephone, cellular phone, cellphone, cell, mobile phone',
                      488: 'chain',
                      489: 'chainlink fence',
                      490: 'chain mail, ring mail, mail, chain armor, chain armour, ring armor, ring armour',
                      491: 'chain saw, chainsaw',
                      492: 'chest',
                      493: 'chiffonier, commode',
                      494: 'chime, bell, gong',
                      495: 'china cabinet, china closet',
                      496: 'Christmas stocking',
                      497: 'church, church building',
                      498: 'cinema, movie theater, movie theatre, movie house, picture palace',
                      499: 'cleaver, meat cleaver, chopper',
                      500: 'cliff dwelling',
                      501: 'cloak',
                      502: 'clog, geta, patten, sabot',
                      503: 'cocktail shaker',
                      504: 'coffee mug',
                      505: 'coffeepot',
                      506: 'coil, spiral, volute, whorl, helix',
                      507: 'combination lock',
                      508: 'computer keyboard, keypad',
                      509: 'confectionery, confectionary, candy store',
                      510: 'container ship, containership, container vessel',
                      511: 'convertible',
                      512: 'corkscrew, bottle screw',
                      513: 'cornet, horn, trumpet, trump',
                      514: 'cowboy boot',
                      515: 'cowboy hat, ten-gallon hat',
                      516: 'cradle',
                      517: 'crane',
                      518: 'crash helmet',
                      519: 'crate',
                      520: 'crib, cot',
                      521: 'Crock Pot',
                      522: 'croquet ball',
                      523: 'crutch',
                      524: 'cuirass',
                      525: 'dam, dike, dyke',
                      526: 'desk',
                      527: 'desktop computer',
                      528: 'dial telephone, dial phone',
                      529: 'diaper, nappy, napkin',
                      530: 'digital clock',
                      531: 'digital watch',
                      532: 'dining table, board',
                      533: 'dishrag, dishcloth',
                      534: 'dishwasher, dish washer, dishwashing machine',
                      535: 'disk brake, disc brake',
                      536: 'dock, dockage, docking facility',
                      537: 'dogsled, dog sled, dog sleigh',
                      538: 'dome',
                      539: 'doormat, welcome mat',
                      540: 'drilling platform, offshore rig',
                      541: 'drum, membranophone, tympan',
                      542: 'drumstick',
                      543: 'dumbbell',
                      544: 'Dutch oven',
                      545: 'electric fan, blower',
                      546: 'electric guitar',
                      547: 'electric locomotive',
                      548: 'entertainment center',
                      549: 'envelope',
                      550: 'espresso maker',
                      551: 'face powder',
                      552: 'feather boa, boa',
                      553: 'file, file cabinet, filing cabinet',
                      554: 'fireboat',
                      555: 'fire engine, fire truck',
                      556: 'fire screen, fireguard',
                      557: 'flagpole, flagstaff',
                      558: 'flute, transverse flute',
                      559: 'folding chair',
                      560: 'football helmet',
                      561: 'forklift',
                      562: 'fountain',
                      563: 'fountain pen',
                      564: 'four-poster',
                      565: 'freight car',
                      566: 'French horn, horn',
                      567: 'frying pan, frypan, skillet',
                      568: 'fur coat',
                      569: 'garbage truck, dustcart',
                      570: 'gasmask, respirator, gas helmet',
                      571: 'gas pump, gasoline pump, petrol pump, island dispenser',
                      572: 'goblet',
                      573: 'go-kart',
                      574: 'golf ball',
                      575: 'golfcart, golf cart',
                      576: 'gondola',
                      577: 'gong, tam-tam',
                      578: 'gown',
                      579: 'grand piano, grand',
                      580: 'greenhouse, nursery, glasshouse',
                      581: 'grille, radiator grille',
                      582: 'grocery store, grocery, food market, market',
                      583: 'guillotine',
                      584: 'hair slide',
                      585: 'hair spray',
                      586: 'half track',
                      587: 'hammer',
                      588: 'hamper',
                      589: 'hand blower, blow dryer, blow drier, hair dryer, hair drier',
                      590: 'hand-held computer, hand-held microcomputer',
                      591: 'handkerchief, hankie, hanky, hankey',
                      592: 'hard disc, hard disk, fixed disk',
                      593: 'harmonica, mouth organ, harp, mouth harp',
                      594: 'harp',
                      595: 'harvester, reaper',
                      596: 'hatchet',
                      597: 'holster',
                      598: 'home theater, home theatre',
                      599: 'honeycomb',
                      600: 'hook, claw',
                      601: 'hoopskirt, crinoline',
                      602: 'horizontal bar, high bar',
                      603: 'horse cart, horse-cart',
                      604: 'hourglass',
                      605: 'iPod',
                      606: 'iron, smoothing iron',
                      607: "jack-o'-lantern",
                      608: 'jean, blue jean, denim',
                      609: 'jeep, landrover',
                      610: 'jersey, T-shirt, tee shirt',
                      611: 'jigsaw puzzle',
                      612: 'jinrikisha, ricksha, rickshaw',
                      613: 'joystick',
                      614: 'kimono',
                      615: 'knee pad',
                      616: 'knot',
                      617: 'lab coat, laboratory coat',
                      618: 'ladle',
                      619: 'lampshade, lamp shade',
                      620: 'laptop, laptop computer',
                      621: 'lawn mower, mower',
                      622: 'lens cap, lens cover',
                      623: 'letter opener, paper knife, paperknife',
                      624: 'library',
                      625: 'lifeboat',
                      626: 'lighter, light, igniter, ignitor',
                      627: 'limousine, limo',
                      628: 'liner, ocean liner',
                      629: 'lipstick, lip rouge',
                      630: 'Loafer',
                      631: 'lotion',
                      632: 'loudspeaker, speaker, speaker unit, loudspeaker system, speaker system',
                      633: "loupe, jeweler's loupe",
                      634: 'lumbermill, sawmill',
                      635: 'magnetic compass',
                      636: 'mailbag, postbag',
                      637: 'mailbox, letter box',
                      638: 'maillot',
                      639: 'maillot, tank suit',
                      640: 'manhole cover',
                      641: 'maraca',
                      642: 'marimba, xylophone',
                      643: 'mask',
                      644: 'matchstick',
                      645: 'maypole',
                      646: 'maze, labyrinth',
                      647: 'measuring cup',
                      648: 'medicine chest, medicine cabinet',
                      649: 'megalith, megalithic structure',
                      650: 'microphone, mike',
                      651: 'microwave, microwave oven',
                      652: 'military uniform',
                      653: 'milk can',
                      654: 'minibus',
                      655: 'miniskirt, mini',
                      656: 'minivan',
                      657: 'missile',
                      658: 'mitten',
                      659: 'mixing bowl',
                      660: 'mobile home, manufactured home',
                      661: 'Model T',
                      662: 'modem',
                      663: 'monastery',
                      664: 'monitor',
                      665: 'moped',
                      666: 'mortar',
                      667: 'mortarboard',
                      668: 'mosque',
                      669: 'mosquito net',
                      670: 'motor scooter, scooter',
                      671: 'mountain bike, all-terrain bike, off-roader',
                      672: 'mountain tent',
                      673: 'mouse, computer mouse',
                      674: 'mousetrap',
                      675: 'moving van',
                      676: 'muzzle',
                      677: 'nail',
                      678: 'neck brace',
                      679: 'necklace',
                      680: 'nipple',
                      681: 'notebook, notebook computer',
                      682: 'obelisk',
                      683: 'oboe, hautboy, hautbois',
                      684: 'ocarina, sweet potato',
                      685: 'odometer, hodometer, mileometer, milometer',
                      686: 'oil filter',
                      687: 'organ, pipe organ',
                      688: 'oscilloscope, scope, cathode-ray oscilloscope, CRO',
                      689: 'overskirt',
                      690: 'oxcart',
                      691: 'oxygen mask',
                      692: 'packet',
                      693: 'paddle, boat paddle',
                      694: 'paddlewheel, paddle wheel',
                      695: 'padlock',
                      696: 'paintbrush',
                      697: "pajama, pyjama, pj's, jammies",
                      698: 'palace',
                      699: 'panpipe, pandean pipe, syrinx',
                      700: 'paper towel',
                      701: 'parachute, chute',
                      702: 'parallel bars, bars',
                      703: 'park bench',
                      704: 'parking meter',
                      705: 'passenger car, coach, carriage',
                      706: 'patio, terrace',
                      707: 'pay-phone, pay-station',
                      708: 'pedestal, plinth, footstall',
                      709: 'pencil box, pencil case',
                      710: 'pencil sharpener',
                      711: 'perfume, essence',
                      712: 'Petri dish',
                      713: 'photocopier',
                      714: 'pick, plectrum, plectron',
                      715: 'pickelhaube',
                      716: 'picket fence, paling',
                      717: 'pickup, pickup truck',
                      718: 'pier',
                      719: 'piggy bank, penny bank',
                      720: 'pill bottle',
                      721: 'pillow',
                      722: 'ping-pong ball',
                      723: 'pinwheel',
                      724: 'pirate, pirate ship',
                      725: 'pitcher, ewer',
                      726: "plane, carpenter's plane, woodworking plane",
                      727: 'planetarium',
                      728: 'plastic bag',
                      729: 'plate rack',
                      730: 'plow, plough',
                      731: "plunger, plumber's helper",
                      732: 'Polaroid camera, Polaroid Land camera',
                      733: 'pole',
                      734: 'police van, police wagon, paddy wagon, patrol wagon, wagon, black Maria',
                      735: 'poncho',
                      736: 'pool table, billiard table, snooker table',
                      737: 'pop bottle, soda bottle',
                      738: 'pot, flowerpot',
                      739: "potter's wheel",
                      740: 'power drill',
                      741: 'prayer rug, prayer mat',
                      742: 'printer',
                      743: 'prison, prison house',
                      744: 'projectile, missile',
                      745: 'projector',
                      746: 'puck, hockey puck',
                      747: 'punching bag, punch bag, punching ball, punchball',
                      748: 'purse',
                      749: 'quill, quill pen',
                      750: 'quilt, comforter, comfort, puff',
                      751: 'racer, race car, racing car',
                      752: 'racket, racquet',
                      753: 'radiator',
                      754: 'radio, wireless',
                      755: 'radio telescope, radio reflector',
                      756: 'rain barrel',
                      757: 'recreational vehicle, RV, R.V.',
                      758: 'reel',
                      759: 'reflex camera',
                      760: 'refrigerator, icebox',
                      761: 'remote control, remote',
                      762: 'restaurant, eating house, eating place, eatery',
                      763: 'revolver, six-gun, six-shooter',
                      764: 'rifle',
                      765: 'rocking chair, rocker',
                      766: 'rotisserie',
                      767: 'rubber eraser, rubber, pencil eraser',
                      768: 'rugby ball',
                      769: 'rule, ruler',
                      770: 'running shoe',
                      771: 'safe',
                      772: 'safety pin',
                      773: 'saltshaker, salt shaker',
                      774: 'sandal',
                      775: 'sarong',
                      776: 'sax, saxophone',
                      777: 'scabbard',
                      778: 'scale, weighing machine',
                      779: 'school bus',
                      780: 'schooner',
                      781: 'scoreboard',
                      782: 'screen, CRT screen',
                      783: 'screw',
                      784: 'screwdriver',
                      785: 'seat belt, seatbelt',
                      786: 'sewing machine',
                      787: 'shield, buckler',
                      788: 'shoe shop, shoe-shop, shoe store',
                      789: 'shoji',
                      790: 'shopping basket',
                      791: 'shopping cart',
                      792: 'shovel',
                      793: 'shower cap',
                      794: 'shower curtain',
                      795: 'ski',
                      796: 'ski mask',
                      797: 'sleeping bag',
                      798: 'slide rule, slipstick',
                      799: 'sliding door',
                      800: 'slot, one-armed bandit',
                      801: 'snorkel',
                      802: 'snowmobile',
                      803: 'snowplow, snowplough',
                      804: 'soap dispenser',
                      805: 'soccer ball',
                      806: 'sock',
                      807: 'solar dish, solar collector, solar furnace',
                      808: 'sombrero',
                      809: 'soup bowl',
                      810: 'space bar',
                      811: 'space heater',
                      812: 'space shuttle',
                      813: 'spatula',
                      814: 'speedboat',
                      815: "spider web, spider's web",
                      816: 'spindle',
                      817: 'sports car, sport car',
                      818: 'spotlight, spot',
                      819: 'stage',
                      820: 'steam locomotive',
                      821: 'steel arch bridge',
                      822: 'steel drum',
                      823: 'stethoscope',
                      824: 'stole',
                      825: 'stone wall',
                      826: 'stopwatch, stop watch',
                      827: 'stove',
                      828: 'strainer',
                      829: 'streetcar, tram, tramcar, trolley, trolley car',
                      830: 'stretcher',
                      831: 'studio couch, day bed',
                      832: 'stupa, tope',
                      833: 'submarine, pigboat, sub, U-boat',
                      834: 'suit, suit of clothes',
                      835: 'sundial',
                      836: 'sunglass',
                      837: 'sunglasses, dark glasses, shades',
                      838: 'sunscreen, sunblock, sun blocker',
                      839: 'suspension bridge',
                      840: 'swab, swob, mop',
                      841: 'sweatshirt',
                      842: 'swimming trunks, bathing trunks',
                      843: 'swing',
                      844: 'switch, electric switch, electrical switch',
                      845: 'syringe',
                      846: 'table lamp',
                      847: 'tank, army tank, armored combat vehicle, armoured combat vehicle',
                      848: 'tape player',
                      849: 'teapot',
                      850: 'teddy, teddy bear',
                      851: 'television, television system',
                      852: 'tennis ball',
                      853: 'thatch, thatched roof',
                      854: 'theater curtain, theatre curtain',
                      855: 'thimble',
                      856: 'thresher, thrasher, threshing machine',
                      857: 'throne',
                      858: 'tile roof',
                      859: 'toaster',
                      860: 'tobacco shop, tobacconist shop, tobacconist',
                      861: 'toilet seat',
                      862: 'torch',
                      863: 'totem pole',
                      864: 'tow truck, tow car, wrecker',
                      865: 'toyshop',
                      866: 'tractor',
                      867: 'trailer truck, tractor trailer, trucking rig, rig, articulated lorry, semi',
                      868: 'tray',
                      869: 'trench coat',
                      870: 'tricycle, trike, velocipede',
                      871: 'trimaran',
                      872: 'tripod',
                      873: 'triumphal arch',
                      874: 'trolleybus, trolley coach, trackless trolley',
                      875: 'trombone',
                      876: 'tub, vat',
                      877: 'turnstile',
                      878: 'typewriter keyboard',
                      879: 'umbrella',
                      880: 'unicycle, monocycle',
                      881: 'upright, upright piano',
                      882: 'vacuum, vacuum cleaner',
                      883: 'vase',
                      884: 'vault',
                      885: 'velvet',
                      886: 'vending machine',
                      887: 'vestment',
                      888: 'viaduct',
                      889: 'violin, fiddle',
                      890: 'volleyball',
                      891: 'waffle iron',
                      892: 'wall clock',
                      893: 'wallet, billfold, notecase, pocketbook',
                      894: 'wardrobe, closet, press',
                      895: 'warplane, military plane',
                      896: 'washbasin, handbasin, washbowl, lavabo, wash-hand basin',
                      897: 'washer, automatic washer, washing machine',
                      898: 'water bottle',
                      899: 'water jug',
                      900: 'water tower',
                      901: 'whiskey jug',
                      902: 'whistle',
                      903: 'wig',
                      904: 'window screen',
                      905: 'window shade',
                      906: 'Windsor tie',
                      907: 'wine bottle',
                      908: 'wing',
                      909: 'wok',
                      910: 'wooden spoon',
                      911: 'wool, woolen, woollen',
                      912: 'worm fence, snake fence, snake-rail fence, Virginia fence',
                      913: 'wreck',
                      914: 'yawl',
                      915: 'yurt',
                      916: 'web site, website, internet site, site',
                      917: 'comic book',
                      918: 'crossword puzzle, crossword',
                      919: 'street sign',
                      920: 'traffic light, traffic signal, stoplight',
                      921: 'book jacket, dust cover, dust jacket, dust wrapper',
                      922: 'menu',
                      923: 'plate',
                      924: 'guacamole',
                      925: 'consomme',
                      926: 'hot pot, hotpot',
                      927: 'trifle',
                      928: 'ice cream, icecream',
                      929: 'ice lolly, lolly, lollipop, popsicle',
                      930: 'French loaf',
                      931: 'bagel, beigel',
                      932: 'pretzel',
                      933: 'cheeseburger',
                      934: 'hotdog, hot dog, red hot',
                      935: 'mashed potato',
                      936: 'head cabbage',
                      937: 'broccoli',
                      938: 'cauliflower',
                      939: 'zucchini, courgette',
                      940: 'spaghetti squash',
                      941: 'acorn squash',
                      942: 'butternut squash',
                      943: 'cucumber, cuke',
                      944: 'artichoke, globe artichoke',
                      945: 'bell pepper',
                      946: 'cardoon',
                      947: 'mushroom',
                      948: 'Granny Smith',
                      949: 'strawberry',
                      950: 'orange',
                      951: 'lemon',
                      952: 'fig',
                      953: 'pineapple, ananas',
                      954: 'banana',
                      955: 'jackfruit, jak, jack',
                      956: 'custard apple',
                      957: 'pomegranate',
                      958: 'hay',
                      959: 'carbonara',
                      960: 'chocolate sauce, chocolate syrup',
                      961: 'dough',
                      962: 'meat loaf, meatloaf',
                      963: 'pizza, pizza pie',
                      964: 'potpie',
                      965: 'burrito',
                      966: 'red wine',
                      967: 'espresso',
                      968: 'cup',
                      969: 'eggnog',
                      970: 'alp',
                      971: 'bubble',
                      972: 'cliff, drop, drop-off',
                      973: 'coral reef',
                      974: 'geyser',
                      975: 'lakeside, lakeshore',
                      976: 'promontory, headland, head, foreland',
                      977: 'sandbar, sand bar',
                      978: 'seashore, coast, seacoast, sea-coast',
                      979: 'valley, vale',
                      980: 'volcano',
                      981: 'ballplayer, baseball player',
                      982: 'groom, bridegroom',
                      983: 'scuba diver',
                      984: 'rapeseed',
                      985: 'daisy',
                      986: "yellow lady's slipper, yellow lady-slipper, Cypripedium calceolus, Cypripedium parviflorum",
                      987: 'corn',
                      988: 'acorn',
                      989: 'hip, rose hip, rosehip',
                      990: 'buckeye, horse chestnut, conker',
                      991: 'coral fungus',
                      992: 'agaric',
                      993: 'gyromitra',
                      994: 'stinkhorn, carrion fungus',
                      995: 'earthstar',
                      996: 'hen-of-the-woods, hen of the woods, Polyporus frondosus, Grifola frondosa',
                      997: 'bolete',
                      998: 'ear, spike, capitulum',
                      999: 'toilet tissue, toilet paper, bathroom tissue'}



import os
import hashlib
from typing import Optional, List

from .www_tools import is_url, get_filename_from_url, download_from_url


class ImageDownloader:
    """
    "Smart" images downloader.
    """

    def __init__(self, article_path: str, skip_list: Optional[List[str]] = None, skip_all_errors: bool = False,
                 img_dir_name: str = 'images', img_public_path: str = '', downloading_timeout: float = -1,
                 deduplication: bool = False):
        self._img_dir_name = img_dir_name
        self._img_public_path = img_public_path
        self._article_file_path = article_path
        self._skip_list = set(skip_list) if skip_list is not None else []
        self._images_dir = os.path.join(os.path.dirname(self._article_file_path), self._img_dir_name)
        self._skip_all_errors = skip_all_errors
        self._downloading_timeout = downloading_timeout if downloading_timeout > 0 else None
        self._deduplication = deduplication

    def download_images(self, images: List[str]) -> dict:
        """
        Download and save images from the list.

        :return URL -> file path mapping.
        """

        replacement_mapping = {}
        hash_to_path_mapping = {}
        skip_list = self._skip_list
        img_count = len(images)
        path_join = os.path.join
        img_dir_name = self._img_dir_name
        img_public_path = self._img_public_path
        images_dir = self._images_dir
        deduplication = self._deduplication

        try:
            os.makedirs(self._images_dir)
        except FileExistsError:
            # Existing directory is not error.
            pass

        for img_num, img_url in enumerate(images):
            assert img_url not in replacement_mapping.keys(), f'BUG: already downloaded image "{img_url}"...'

            if img_url in skip_list:
                print(f'Image {img_num + 1} ["{img_url}"] was skipped, because it\'s in the skip list...')
                continue

            if not is_url(img_url):
                print(f'Image {img_num + 1} ["{img_url}"] was skipped, because it has incorrect URL...')
                continue

            print(f'Downloading image {img_num + 1} of {img_count} from "{img_url}"...')

            try:
                img_response = download_from_url(img_url, self._downloading_timeout)
            except Exception as e:
                if self._skip_all_errors:
                    print(f'Warning: can\'t download image {img_num + 1}, error: [{str(e)}], '
                          'but processing will be continued, because `skip_all_errors` flag is set')
                    continue
                raise

            img_filename = get_filename_from_url(img_response)
            image_content = img_response.content

            if deduplication:
                new_content_hash = hashlib.sha256(image_content).digest()
                existed_file_name = hash_to_path_mapping.get(new_content_hash)
                if existed_file_name is not None:
                    img_filename = existed_file_name
                    document_img_path = path_join(img_public_path or img_dir_name, img_filename)
                    replacement_mapping.setdefault(img_url, document_img_path)
                    continue
                else:
                    hash_to_path_mapping[new_content_hash] = img_filename

            document_img_path = path_join(img_public_path or img_dir_name, img_filename)
            img_filename, document_img_path = self._correct_paths(replacement_mapping, document_img_path, img_url,
                                                                  img_filename)

            real_img_path = path_join(images_dir, img_filename)
            replacement_mapping.setdefault(img_url, document_img_path)

            ImageDownloader._write_image(real_img_path, image_content)

        return replacement_mapping

    @staticmethod
    def _write_image(img_path: str, data: bytes):
        """
        Write image data into the file.
        """

        print(f'Image will be written to the file "{img_path}"...')
        with open(img_path, 'wb') as img_file:
            img_file.write(data)
            img_file.close()

    def _correct_paths(self, replacement_mapping, document_img_path, img_url, img_filename):
        """
        Fix path if a file with the similar name exists already.
        """

        # Images can have similar name, but different URLs, but I want to save original filename, if possible.
        for url, path in replacement_mapping.items():
            if document_img_path == path and img_url != url:
                img_filename = f'{hashlib.md5(img_url.encode()).hexdigest()}_{img_filename}'
                document_img_path = os.path.join(self._img_public_path or self._img_dir_name, img_filename)
                break

        return img_filename, document_img_path





"""

Converting standared Yolo format to VOC format
here

"""


# Script to convert yolo annotations to voc format

# Sample format
# <annotation>
#     <folder>_image_fashion</folder>
#     <filename>brooke-cagle-39574.jpg</filename>
#     <size>
#         <width>1200</width>
#         <height>800</height>
#         <depth>3</depth>
#     </size>
#     <segmented>0</segmented>
#     <object>
#         <name>head</name>
#         <pose>Unspecified</pose>
#         <truncated>0</truncated>
#         <difficult>0</difficult>
#         <bndbox>
#             <xmin>549</xmin>
#             <ymin>251</ymin>
#             <xmax>625</xmax>
#             <ymax>335</ymax>
#         </bndbox>
#     </object>
# <annotation>
import os
import xml.etree.cElementTree as ET
from PIL import Image
import sys
import glob


CLASS_MAPPING = {
    '0': 'name'
    # Add your remaining classes here.
}


def create_root(file_prefix, width, height):
    root = ET.Element("annotations")
    ET.SubElement(root, "filename").text = "{}.jpg".format(file_prefix)
    ET.SubElement(root, "folder").text = "images"
    size = ET.SubElement(root, "size")
    ET.SubElement(size, "width").text = str(width)
    ET.SubElement(size, "height").text = str(height)
    ET.SubElement(size, "depth").text = "3"
    return root


def create_object_annotation(root, voc_labels):
    for voc_label in voc_labels:
        obj = ET.SubElement(root, "object")
        ET.SubElement(obj, "name").text = voc_label[0]
        ET.SubElement(obj, "pose").text = "Unspecified"
        ET.SubElement(obj, "truncated").text = str(0)
        ET.SubElement(obj, "difficult").text = str(0)
        bbox = ET.SubElement(obj, "bndbox")
        ET.SubElement(bbox, "xmin").text = str(voc_label[1])
        ET.SubElement(bbox, "ymin").text = str(voc_label[2])
        ET.SubElement(bbox, "xmax").text = str(voc_label[3])
        ET.SubElement(bbox, "ymax").text = str(voc_label[4])
    return root


def create_file(file_prefix, width, height, voc_labels, des_dir):
    root = create_root(file_prefix, width, height)
    root = create_object_annotation(root, voc_labels)
    tree = ET.ElementTree(root)
    tree.write("{}/{}.xml".format(des_dir, file_prefix))


def read_file(file_path, des_dir, img_dir, classes_names):
    file_prefix = os.path.basename(file_path).split(".txt")[0]
    image_file_name = "{}.jpg".format(file_prefix)
    img_p = os.path.join(img_dir, image_file_name)
    assert os.path.exists(
        img_p), 'make sure all images under: {}'.format(img_dir)
    img = Image.open(img_p)

    w, h = img.size
    with open(file_path, 'r') as file:
        lines = file.readlines()
        voc_labels = []
        for line in lines:
            voc = []
            line = line.strip()
            data = line.split()
            # voc.append(CLASS_MAPPING.get(data[0]))
            if classes_names:
                voc.append(classes_names[int(data[0])])
            else:
                voc.append(data[0])
            bbox_width = float(data[3]) * w
            bbox_height = float(data[4]) * h
            cx = float(data[1]) * w
            cy = float(data[2]) * h
            x = cx - bbox_width/2
            y = cy - bbox_height/2
            voc.append(x)
            voc.append(y)
            voc.append(x + (bbox_width))
            voc.append(y + (bbox_height))
            voc_labels.append(voc)
        create_file(file_prefix, w, h, voc_labels, des_dir)
    print("Processing complete for file: {}".format(file_path))


def yolo2voc(img_dir, txt_dir, class_txt):
    # TODO: need apply a map to mapping yolo label back
    classes_names = None
    if class_txt:
        classes_names = [i.strip() for i in open(class_txt, 'r').readlines()]
    des_d = 'Annotations_yolo_converted_for_voc'
    des_d = os.path.join(os.path.dirname(txt_dir.rstrip('/')), des_d)
    print('VOC Annotations will saved into: {}'.format(des_d))
    os.makedirs(des_d, exist_ok=True)
    txts = glob.glob(os.path.join(txt_dir.rstrip('/'), '*.txt'))
    print('found {} text files in yolo format.'.format(len(txts)))
    for filename in txts:
        read_file(filename, des_d, img_dir, classes_names)


if __name__ == "__main__":
    yolo2voc(sys.argv[1], sys.argv[2], None)



import os
import json
import torch

__all__ = ["get_gpu_prop", "collect_gpu_info"]


dirname = os.path.dirname(__file__)


def get_gpu_devices_count():
    return torch.cuda.device_count()


def get_gpu_prop(show=True):
    ngpus = torch.cuda.device_count()

    properties = []
    for dev in range(ngpus):
        prop = torch.cuda.get_device_properties(dev)
        properties.append({
            "name": prop.name,
            "capability": [prop.major, prop.minor],
            # unit GB
            "total_momory": round(prop.total_memory / 1073741824, 2),
            "sm_count": prop.multi_processor_count
        })

    if show:
        print("cuda: {}".format(torch.cuda.is_available()))
        print("available GPU(s): {}".format(ngpus))
        for i, p in enumerate(properties):
            print("{}: {}".format(i, p))
    return properties


def sort(d, tmp={}):
    for k in sorted(d.keys()):
        if isinstance(d[k], dict):
            tmp[k] = {}
            sort(d[k], tmp[k])
        else:
            tmp[k] = d[k]
    return tmp



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""

Convert quaternion to euler angles


"""
import numpy as np
import math


def euler_to_quaternion(yaw, pitch, roll):
    """
    return a list of [qx, qy, qz, qw]
    """
    qx = np.sin(roll/2) * np.cos(pitch/2) * np.cos(yaw/2) - \
        np.cos(roll/2) * np.sin(pitch/2) * np.sin(yaw/2)
    qy = np.cos(roll/2) * np.sin(pitch/2) * np.cos(yaw/2) + \
        np.sin(roll/2) * np.cos(pitch/2) * np.sin(yaw/2)
    qz = np.cos(roll/2) * np.cos(pitch/2) * np.sin(yaw/2) - \
        np.sin(roll/2) * np.sin(pitch/2) * np.cos(yaw/2)
    qw = np.cos(roll/2) * np.cos(pitch/2) * np.cos(yaw/2) + \
        np.sin(roll/2) * np.sin(pitch/2) * np.sin(yaw/2)
    return [qx, qy, qz, qw]


def quaternion_to_euler(x, y, z, w):
    """
    return a list of euler angles [yaw, pitch, roll]
    """
    t0 = +2.0 * (w * x + y * z)
    t1 = +1.0 - 2.0 * (x * x + y * y)
    roll = math.atan2(t0, t1)
    t2 = +2.0 * (w * y - z * x)
    t2 = +1.0 if t2 > +1.0 else t2
    t2 = -1.0 if t2 < -1.0 else t2
    pitch = math.asin(t2)
    t3 = +2.0 * (w * z + x * y)
    t4 = +1.0 - 2.0 * (y * y + z * z)
    yaw = math.atan2(t3, t4)
    return [yaw, pitch, roll]



from .sliceable_dataset import _as_key_indices
from .sliceable_dataset import _is_iterable
from .sliceable_dataset import SliceableDataset


class GetterDataset(SliceableDataset):
    """A sliceable dataset class that is defined with getters.

    This is a dataset class with getters.
    Please refer to the tutorial for more detailed explanation.

    Here is an example.

    >>> class SliceableLabeledImageDataset(GetterDataset):
    >>>     def __init__(self, pairs, root='.'):
    >>>         super(SliceableLabeledImageDataset, self).__init__()
    >>>         with open(pairs) as f:
    >>>             self._pairs = [l.split() for l in f]
    >>>         self._root = root
    >>>
    >>>         self.add_getter('img', self.get_image)
    >>>         self.add_getter('label', self.get_label)
    >>>
    >>>     def __len__(self):
    >>>         return len(self._pairs)
    >>>
    >>>     def get_image(self, i):
    >>>         path, _ = self._pairs[i]
    >>>         return read_image(os.path.join(self._root, path))
    >>>
    >>>     def get_label(self, i):
    >>>         _, label = self._pairs[i]
    >>>         return np.int32(label)
    >>>
    >>> dataset = SliceableLabeledImageDataset('list.txt')
    >>>
    >>> # get a subset with label = 0, 1, 2
    >>> # no images are loaded
    >>> indices = [i for i, label in
    ...            enumerate(dataset.slice[:, 'label']) if label in {0, 1, 2}]
    >>> dataset_012 = dataset.slice[indices]
    """

    def __init__(self):
        self._keys = []
        self._getters = []
        self._return_tuple = True

    def __len__(self):
        raise NotImplementedError

    @property
    def keys(self):
        if self._return_tuple:
            return tuple(key for key, _, _ in self._keys)
        else:
            return self._keys[0][0]

    @keys.setter
    def keys(self, keys):
        self._keys = [self._keys[key_index]
                      for key_index in _as_key_indices(keys, self.keys)]
        self._return_tuple = _is_iterable(keys)

    def add_getter(self, keys, getter):
        """Register a getter function

        Args:
            keys (string or tuple of strings): The name(s) of data
                that the getter function returns.
            getter (callable): A getter function that takes an index and
                returns data of the corresponding example.
        """
        self._getters.append(getter)
        if _is_iterable(keys):
            for key_index, key in enumerate(keys):
                self._keys.append((key, len(self._getters) - 1, key_index))
        else:
            self._keys.append((keys, len(self._getters) - 1, None))

    def get_example_by_keys(self, index, key_indices):
        example = []
        cache = {}
        for key_index in key_indices:
            _, getter_index, key_index = self._keys[key_index]
            if getter_index not in cache:
                cache[getter_index] = self._getters[getter_index](index)
            if key_index is None:
                example.append(cache[getter_index])
            else:
                example.append(cache[getter_index][key_index])
        return tuple(example)



#!/usr/bin/env python
#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
Given a CVAT XML and a directory with the image dataset, this script reads the
CVAT XML and writes the annotations in PASCAL VOC format into a given
directory.

This implementation supports both interpolation tracks from video and 
annotated images.  If it encounters any tracks or annotations that are 
not bounding boxes, it ignores them.
"""

import os
import argparse
from alfred.utils.log import logger as log
from lxml import etree
from pascal_voc_writer import Writer
from collections import OrderedDict


def parse_args():
    """Parse arguments of command line"""
    parser = argparse.ArgumentParser(
        description='Convert CVAT XML annotations to PASCAL VOC format'
    )

    parser.add_argument(
        '--cvat-xml', metavar='FILE', required=True,
        help='input file with CVAT annotation in xml format'
    )

    parser.add_argument(
        '--image-dir', metavar='DIRECTORY', required=True,
        help='directory which contains original images'
    )

    parser.add_argument(
        '--output-dir', metavar='DIRECTORY', required=True,
        help='directory for output annotations in PASCAL VOC format'
    )

    return parser.parse_args()


def process_cvat_xml(xml_file, image_dir, output_dir):
    """
    Transforms a single XML in CVAT format to multiple PASCAL VOC format
    XMls.

    :param xml_file: CVAT format XML
    :param image_dir: image directory of the dataset
    :param output_dir: directory of annotations with PASCAL VOC format
    :return:
    """
    KNOWN_TAGS = {'box', 'image', 'attribute'}
    os.makedirs(output_dir, exist_ok=True)
    cvat_xml = etree.parse(xml_file)

    basename = os.path.splitext( os.path.basename( xml_file ) )[0]

    tracks= cvat_xml.findall( './/track' )
    log.info('tracks: {}'.format(tracks))

    if (tracks is not None) and (len(tracks) > 0):
        frames = {}

        for track in tracks:
            trackid = int(track.get("id"))
            label = track.get("label")
            boxes = track.findall( './box' )
            for box in boxes:
                frameid  = int(box.get('frame'))
                outside  = int(box.get('outside'))
                #occluded = int(box.get('occluded'))  #currently unused
                #keyframe = int(box.get('keyframe'))  #currently unused
                xtl      = float(box.get('xtl'))
                ytl      = float(box.get('ytl'))
                xbr      = float(box.get('xbr'))
                ybr      = float(box.get('ybr'))
                
                frame = frames.get( frameid, {} )
                
                if outside == 0:
                    frame[ trackid ] = { 'xtl': xtl, 'ytl': ytl, 'xbr': xbr, 'ybr': ybr, 'label': label }

                frames[ frameid ] = frame

        width = int(cvat_xml.find('.//original_size/width').text)
        height  = int(cvat_xml.find('.//original_size/height').text)

        # Spit out a list of each object for each frame
        for frameid in sorted(frames.keys()):
            #print( frameid )

            image_name = "%s_%08d.jpg" % (basename, frameid)
            image_path = os.path.join(image_dir, image_name)
            if not os.path.exists(image_path):
                log.info('{} image cannot be found. Is `{}` image directory correct?'.
                    format(image_path, image_dir))
            writer = Writer(image_path, width, height)

            frame = frames[frameid]

            objids = sorted(frame.keys())

            for objid in objids:

                box = frame[objid]

                label = box.get('label')
                xmin = float(box.get('xtl'))
                ymin = float(box.get('ytl'))
                xmax = float(box.get('xbr'))
                ymax = float(box.get('ybr'))

                writer.addObject(label, xmin, ymin, xmax, ymax)

            anno_name = os.path.basename(os.path.splitext(image_name)[0] + '.xml')
            anno_dir = os.path.dirname(os.path.join(output_dir, image_name))
            os.makedirs(anno_dir, exist_ok=True)
            writer.save(os.path.join(anno_dir, anno_name))

    else:
        for img_tag in cvat_xml.findall('image'):
            image_name = img_tag.get('name')
            width = img_tag.get('width')
            height = img_tag.get('height')
            depth = img_tag.get('depth', 3)
            image_path = os.path.join(image_dir, image_name)
            if not os.path.exists(image_path):
                log.info('{} image cannot be found. Is `{}` image directory correct?'.
                    format(image_path, image_dir))
            writer = Writer(image_path, width, height, depth=depth)

            unknown_tags = {x.tag for x in img_tag.iter()}.difference(KNOWN_TAGS)
            if unknown_tags:
                log.info('Ignoring tags for image {}: {}'.format(image_path, unknown_tags))

            for box in img_tag.findall('box'):
                label = box.get('label')
                # concat label with attributes
                # todo: check if exist or not
                all_attributes = box.findall('attribute')
                attr_dict = OrderedDict()
                for attr in all_attributes:
                    attr_dict[attr.get('name')] = attr.text
                lst = sorted(attr_dict.items(), key=lambda item: item[0])
                attr_dict = OrderedDict(lst)
                # label += '_' + '_'.join(attr_dict.values())
                # we only take color for now
                label = label.replace('_', '')
                label += '_' + list(attr_dict.values())[0]
                # log.info(label)
                xmin = float(box.get('xtl'))
                ymin = float(box.get('ytl'))
                xmax = float(box.get('xbr'))
                ymax = float(box.get('ybr'))

                writer.addObject(label, xmin, ymin, xmax, ymax)

            anno_name = os.path.basename(os.path.splitext(image_name)[0] + '.xml')
            anno_dir = os.path.dirname(os.path.join(output_dir, image_name))
            os.makedirs(anno_dir, exist_ok=True)
            writer.save(os.path.join(anno_dir, anno_name))


def main():
    args = parse_args()
    process_cvat_xml(args.cvat_xml, args.image_dir, args.output_dir)


if __name__ == "__main__":
    main()


import cv2
import numpy as np


def draw_face_landmarks(img, pts, box=None, color=(255, 147, 23), size=8):
    if pts is not None:
        print(pts.shape)
        n = pts.shape[1]
        if n <= 106:
            for i in range(n):
                cv2.circle(
                    img, (int(round(pts[0, i])), int(round(pts[1, i]))), size, color, -1
                )
        else:
            sep = 1
            for i in range(0, n, sep):
                cv2.circle(
                    img, (int(round(pts[0, i])), int(round(pts[1, i]))), 2, color, 1
                )

        if box is not None:
            line_color = (255, 127, 80)
            left, top, right, bottom = np.round(box).astype(np.int32)
            left_top = (left, top)
            right_top = (right, top)
            right_bottom = (right, bottom)
            left_bottom = (left, bottom)
            cv2.line(img, left_top, right_top, line_color, 1, cv2.LINE_AA)
            cv2.line(img, right_top, right_bottom, line_color, 1, cv2.LINE_AA)
            cv2.line(img, right_bottom, left_bottom, line_color, 1, cv2.LINE_AA)
            cv2.line(img, left_bottom, left_top, line_color, 1, cv2.LINE_AA)
    return img




#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""

this tool helps viewing VOC format data


"""
import os
import sys
import cv2
import xml.etree.ElementTree as ET
from glob import glob
import os
import sys
import cv2
from alfred.utils.log import logger as logging
from pascal_voc_writer import Writer
from lxml import etree


def get_and_check(root, name, length):
    vars = root.findall(name)
    if len(vars) == 0:
        raise NotImplementedError('Can not find %s in %s.' % (name, root.tag))
    if length > 0 and len(vars) != length:
        raise NotImplementedError('The size of %s is supposed to be %d, but is %d.' % (name, length, len(vars)))
    if length == 1:
        vars = vars[0]
    return vars


def get(root, name):
    vars = root.findall(name)
    return vars

def merge_multiple_voc_labels(n_merge_files, target_save_root):
    # one of n_merge_files must exist, but no sure which one, found it out
    f_name = os.path.basename(n_merge_files[0])
    n = len(n_merge_files)
    for f in n_merge_files:
        if os.path.exists(f):
            tree = ET.parse(f)
            root = tree.getroot()
            objs = get(root, 'object')
            n_merge_files.remove(f)
            break
    for xml in n_merge_files:
        if os.path.exists(xml):
            t = ET.parse(xml)
            r = t.getroot()
            for obj in get(r, 'object'):
                root.append(obj)
    # how to save 

    # f = open(os.path.join(target_save_root, f_name), 'wb')
    # f.write(etree.tostring(root, pretty_print=True))
    # f.close()
    tree.write(os.path.join(target_save_root, f_name))
    print('merged {} xmls and saved into: {}'.format(n, os.path.join(target_save_root, f_name)))


def merge_voc(label_root_list, style='intersection', label_major=True):
    """
    For intersection: only merges their intersection part;
    For union: will merge all of them

    merge VOC with multiple datasets (one of them may have partial object labeled)
    """
    logging.info('labels root: {}'.format(label_root_list))
    # these labels may not perfectly aligned, we get a union of them
    filenames = []
    if style == 'union':
        logging.info('merge with union style.')
        for l in label_root_list:
            xmls = [os.path.basename(i) for i in glob(os.path.join(l, "*.xml"))]
            logging.info('found {} xmls under: {}'.format(len(xmls), l))
            # filenames.extend([i for i in xmls if i not in filenames])
            filenames.extend(xmls)
    else:
        logging.info('merge with intersection style.')
        for l in label_root_list:
            xmls = [os.path.basename(i) for i in glob(os.path.join(l, "*.xml"))]
            logging.info('found {} xmls under: {}'.format(len(xmls), l))
            # filenames.extend([i for i in xmls if i not in filenames])
            if len(filenames) > 0:
                filenames = set(xmls) & set(filenames)
            filenames = xmls

    filenames = list(set(filenames))
    logging.info('found {} xmls which exist both inside all provided label roots.'.format(len(filenames)))
    target_save_root = './merged_voc_annotations'
    os.makedirs(target_save_root, exist_ok=True)
    for f in filenames:
        n_merge_files = []
        for l in label_root_list:
            n_merge_files.append(os.path.join(l, f))
        merge_multiple_voc_labels(n_merge_files, target_save_root)
    print('done.')


if __name__ == "__main__":
    merge_voc(sys.argv[1], sys.argv[2])



"""
Routines for the strings.
"""

import re
import unicodedata


def slugify(value):
    """
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.
    """

    value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore')
    value = re.sub(r'[^\w\s-]', '', value.decode()).strip().lower()
    value = re.sub(r'[-\s]+', '-', value)

    return value



import numpy
import six

# below code from Chianer

class DatasetMixin(object):

    """Default implementation of dataset indexing.

    DatasetMixin provides the :meth:`__getitem__` operator. The default
    implementation uses :meth:`get_example` to extract each example, and
    combines the results into a list. This mixin makes it easy to implement a
    new dataset that does not support efficient slicing.

    Dataset implementation using DatasetMixin still has to provide the
    :meth:`__len__` operator explicitly.

    """

    def __getitem__(self, index):
        """Returns an example or a sequence of examples.

        It implements the standard Python indexing and one-dimensional integer
        array indexing. It uses the :meth:`get_example` method by default, but
        it may be overridden by the implementation to, for example, improve the
        slicing performance.

        Args:
            index (int, slice, list or numpy.ndarray): An index of an example
                or indexes of examples.

        Returns:
            If index is int, returns an example created by `get_example`.
            If index is either slice or one-dimensional list or numpy.ndarray,
            returns a list of examples created by `get_example`.

        .. admonition:: Example

           >>> import numpy
           >>> from chainer import dataset
           >>> class SimpleDataset(dataset.DatasetMixin):
           ...     def __init__(self, values):
           ...         self.values = values
           ...     def __len__(self):
           ...         return len(self.values)
           ...     def get_example(self, i):
           ...         return self.values[i]
           ...
           >>> ds = SimpleDataset([0, 1, 2, 3, 4, 5])
           >>> ds[1]   # Access by int
           1
           >>> ds[1:3]  # Access by slice
           [1, 2]
           >>> ds[[4, 0]]  # Access by one-dimensional integer list
           [4, 0]
           >>> index = numpy.arange(3)
           >>> ds[index]  # Access by one-dimensional integer numpy.ndarray
           [0, 1, 2]

        """
        if isinstance(index, slice):
            current, stop, step = index.indices(len(self))
            return [self.get_example(i) for i in
                    six.moves.range(current, stop, step)]
        elif isinstance(index, list) or isinstance(index, numpy.ndarray):
            return [self.get_example(i) for i in index]
        else:
            return self.get_example(index)

    def __len__(self):
        """Returns the number of data points."""
        raise NotImplementedError

    def get_example(self, i):
        """Returns the i-th example.

        Implementations should override it. It should raise :class:`IndexError`
        if the index is invalid.

        Args:
            i (int): The index of the example.

        Returns:
            The i-th example.

        """
        raise NotImplementedError



import functools
import torch
import torch.nn.functional as F


def bbox_overlaps(bboxes1, bboxes2, mode='iou', is_aligned=False, eps=1e-6):
    """Calculate overlap between two set of bboxes.

    If ``is_aligned `` is ``False``, then calculate the overlaps between each
    bbox of bboxes1 and bboxes2, otherwise the overlaps between each aligned
    pair of bboxes1 and bboxes2.

    Args:
        bboxes1 (Tensor): shape (B, m, 4) in <x1, y1, x2, y2> format or empty.
        bboxes2 (Tensor): shape (B, n, 4) in <x1, y1, x2, y2> format or empty.
            B indicates the batch dim, in shape (B1, B2, ..., Bn).
            If ``is_aligned `` is ``True``, then m and n must be equal.
        mode (str): "iou" (intersection over union) or "iof" (intersection over
            foreground).
        is_aligned (bool, optional): If True, then m and n must be equal.
            Default False.
        eps (float, optional): A value added to the denominator for numerical
            stability. Default 1e-6.

    Returns:
        Tensor: shape (m, n) if ``is_aligned `` is False else shape (m,)

    Example:
        >>> bboxes1 = torch.FloatTensor([
        >>>     [0, 0, 10, 10],
        >>>     [10, 10, 20, 20],
        >>>     [32, 32, 38, 42],
        >>> ])
        >>> bboxes2 = torch.FloatTensor([
        >>>     [0, 0, 10, 20],
        >>>     [0, 10, 10, 19],
        >>>     [10, 10, 20, 20],
        >>> ])
        >>> bbox_overlaps(bboxes1, bboxes2)
        tensor([[0.5000, 0.0000, 0.0000],
                [0.0000, 0.0000, 1.0000],
                [0.0000, 0.0000, 0.0000]])
        >>> bbox_overlaps(bboxes1, bboxes2, mode='giou', eps=1e-7)
        tensor([[0.5000, 0.0000, -0.5000],
                [-0.2500, -0.0500, 1.0000],
                [-0.8371, -0.8766, -0.8214]])

    Example:
        >>> empty = torch.FloatTensor([])
        >>> nonempty = torch.FloatTensor([
        >>>     [0, 0, 10, 9],
        >>> ])
        >>> assert tuple(bbox_overlaps(empty, nonempty).shape) == (0, 1)
        >>> assert tuple(bbox_overlaps(nonempty, empty).shape) == (1, 0)
        >>> assert tuple(bbox_overlaps(empty, empty).shape) == (0, 0)
    """

    assert mode in ['iou', 'iof', 'giou'], f'Unsupported mode {mode}'
    # Either the boxes are empty or the length of boxes's last dimenstion is 4
    assert (bboxes1.size(-1) == 4 or bboxes1.size(0) == 0)
    assert (bboxes2.size(-1) == 4 or bboxes2.size(0) == 0)

    # Batch dim must be the same
    # Batch dim: (B1, B2, ... Bn)
    assert bboxes1.shape[:-2] == bboxes2.shape[:-2]
    batch_shape = bboxes1.shape[:-2]

    rows = bboxes1.size(-2)
    cols = bboxes2.size(-2)
    if is_aligned:
        assert rows == cols

    if rows * cols == 0:
        if is_aligned:
            return bboxes1.new(batch_shape + (rows, ))
        else:
            return bboxes1.new(batch_shape + (rows, cols))

    area1 = (bboxes1[..., 2] - bboxes1[..., 0]) * (
        bboxes1[..., 3] - bboxes1[..., 1])
    area2 = (bboxes2[..., 2] - bboxes2[..., 0]) * (
        bboxes2[..., 3] - bboxes2[..., 1])

    if is_aligned:
        lt = torch.max(bboxes1[..., :2], bboxes2[..., :2])  # [B, rows, 2]
        rb = torch.min(bboxes1[..., 2:], bboxes2[..., 2:])  # [B, rows, 2]

        wh = (rb - lt).clamp(min=0)  # [B, rows, 2]
        overlap = wh[..., 0] * wh[..., 1]

        if mode in ['iou', 'giou']:
            union = area1 + area2 - overlap
        else:
            union = area1
        if mode == 'giou':
            enclosed_lt = torch.min(bboxes1[..., :2], bboxes2[..., :2])
            enclosed_rb = torch.max(bboxes1[..., 2:], bboxes2[..., 2:])
    else:
        lt = torch.max(bboxes1[..., :, None, :2],
                       bboxes2[..., None, :, :2])  # [B, rows, cols, 2]
        rb = torch.min(bboxes1[..., :, None, 2:],
                       bboxes2[..., None, :, 2:])  # [B, rows, cols, 2]

        wh = (rb - lt).clamp(min=0)  # [B, rows, cols, 2]
        overlap = wh[..., 0] * wh[..., 1]

        if mode in ['iou', 'giou']:
            union = area1[..., None] + area2[..., None, :] - overlap
        else:
            union = area1[..., None]
        if mode == 'giou':
            enclosed_lt = torch.min(bboxes1[..., :, None, :2],
                                    bboxes2[..., None, :, :2])
            enclosed_rb = torch.max(bboxes1[..., :, None, 2:],
                                    bboxes2[..., None, :, 2:])

    eps = union.new_tensor([eps])
    union = torch.max(union, eps)
    ious = overlap / union
    if mode in ['iou', 'iof']:
        return ious
    # calculate gious
    enclose_wh = (enclosed_rb - enclosed_lt).clamp(min=0)
    enclose_area = enclose_wh[..., 0] * enclose_wh[..., 1]
    enclose_area = torch.max(enclose_area, eps)
    gious = ious - (enclose_area - union) / enclose_area
    return gious


def reduce_loss(loss, reduction):
    """Reduce loss as specified.

    Args:
        loss (Tensor): Elementwise loss tensor.
        reduction (str): Options are "none", "mean" and "sum".

    Return:
        Tensor: Reduced loss tensor.
    """
    reduction_enum = F._Reduction.get_enum(reduction)
    # none: 0, elementwise_mean:1, sum: 2
    if reduction_enum == 0:
        return loss
    elif reduction_enum == 1:
        return loss.mean()
    elif reduction_enum == 2:
        return loss.sum()


def weight_reduce_loss(loss, weight=None, reduction='mean', avg_factor=None):
    """Apply element-wise weight and reduce loss.

    Args:
        loss (Tensor): Element-wise loss.
        weight (Tensor): Element-wise weights.
        reduction (str): Same as built-in losses of PyTorch.
        avg_factor (float): Avarage factor when computing the mean of losses.

    Returns:
        Tensor: Processed loss values.
    """
    # if weight is specified, apply element-wise weight
    if weight is not None:
        loss = loss * weight

    # if avg_factor is not specified, just reduce the loss
    if avg_factor is None:
        loss = reduce_loss(loss, reduction)
    else:
        # if reduction is mean, then average the loss by avg_factor
        if reduction == 'mean':
            loss = loss.sum() / avg_factor
        # if reduction is 'none', then do nothing, otherwise raise an error
        elif reduction != 'none':
            raise ValueError('avg_factor can not be used with reduction="sum"')
    return loss


def weighted_loss(loss_func):
    """Create a weighted version of a given loss function.

    To use this decorator, the loss function must have the signature like
    `loss_func(pred, target, **kwargs)`. The function only needs to compute
    element-wise loss without any reduction. This decorator will add weight
    and reduction arguments to the function. The decorated function will have
    the signature like `loss_func(pred, target, weight=None, reduction='mean',
    avg_factor=None, **kwargs)`.

    :Example:

    >>> import torch
    >>> @weighted_loss
    >>> def l1_loss(pred, target):
    >>>     return (pred - target).abs()

    >>> pred = torch.Tensor([0, 2, 3])
    >>> target = torch.Tensor([1, 1, 1])
    >>> weight = torch.Tensor([1, 0, 1])

    >>> l1_loss(pred, target)
    tensor(1.3333)
    >>> l1_loss(pred, target, weight)
    tensor(1.)
    >>> l1_loss(pred, target, reduction='none')
    tensor([1., 1., 2.])
    >>> l1_loss(pred, target, weight, avg_factor=2)
    tensor(1.5000)
    """

    @functools.wraps(loss_func)
    def wrapper(pred,
                target,
                weight=None,
                reduction='mean',
                avg_factor=None,
                **kwargs):
        # get element-wise loss
        loss = loss_func(pred, target, **kwargs)
        loss = weight_reduce_loss(loss, weight, reduction, avg_factor)
        return loss

    return wrapper



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
Common utility of pytorch

this contains code that frequently used while writing torch applications

"""
import itertools
import inspect
from json.tool import main
import torch
from colorama import Fore, Back, Style
import numpy as np


device = torch.device(
    'cuda') if torch.cuda.is_available() else torch.device('cpu')


def print_tensor(t, label=None, ignore_value=True):
    if isinstance(t, torch.Tensor):
        if label:
            print(Fore.YELLOW + Style.BRIGHT +
                  "-> {}".format(label) + Style.RESET_ALL)
        else:
            print(Fore.YELLOW + Style.BRIGHT + "tensor: " + Style.RESET_ALL)
        if ignore_value:
            print('shape: {}\ndtype: {} {}\n'.format(
                t.shape, t.dtype, t.device))
        else:
            print('value: {}\nshape: {}\ndtype: {}\n'.format(
                t, t.shape, t.dtype
            ))

    else:
        print('{} is not a tensor.'.format(t))


def decorator(f):
    def wrapper(*args, **kwargs):
        bound_args = inspect.signature(f).bind(*args, **kwargs)
        bound_args.apply_defaults()

        frame = inspect.currentframe()
        frame = inspect.getouterframes(frame)[1]
        string = inspect.getframeinfo(frame[0]).code_context[0].strip()
        args_ori_names = string[string.find('(') + 1:-1].split(',')

        names = []
        for i in args_ori_names:
            if i.find('=') != -1:
                names.append(i.split('=')[1].strip())
            else:
                names.append(i)
        args_dict = dict(zip(names, args))
        for k, v in args_dict.items():
            k = k.strip()
            if isinstance(v, torch.Tensor):
                print(f'[{k}]: ', v.shape, v.device, v.dtype)
            else:
                print(f'[{k}]: ', v.shape)
        return f(*args, **kwargs)
    return wrapper


@decorator
def print_shape(*vs):
    pass


if __name__ == '__main__':
    cam = torch.randn([4, 5, 300])
    pose = torch.randn([1, 44, 55])
    # print_shape(locals(), cam, pose)
    print_shape(cam, pose)



#!/usr/bin/python3

"""
Simple script to download images and replace image links in markdown documents.
"""

import argparse
import os

from time import strftime
from mimetypes import types_map

from .mdparse.transformers.md.transformer import ArticleTransformer
from .mdparse.image_downloader import ImageDownloader
from .mdparse.www_tools import is_url, get_filename_from_url, download_from_url
from .mdparse.formatters.simple import SimpleFormatter
from .mdparse.formatters.html import HTMLFormatter
try:
    from .mdparse.formatters.pdf import PDFFormatter
except ModuleNotFoundError:
    PDFFormatter = None


__version__ = '0.0.4'
FORMATTERS = [SimpleFormatter, HTMLFormatter, PDFFormatter]
del types_map['.jpe']


def download_images_from_md(md_f):
    article_link = md_f
    if is_url(article_link):
        response = download_from_url(
            article_link, timeout=-1)
        article_path = get_filename_from_url(response)

        with open(article_path, 'wb') as article_file:
            article_file.write(response.content)
            article_file.close()
    else:
        article_path = os.path.expanduser(article_link)

    img_dir_name = os.path.basename(article_path).split('.')[0]
    img_dir_name = os.path.join('images', img_dir_name)
    os.makedirs(img_dir_name, exist_ok=True)

    img_downloader = ImageDownloader(
        article_path=article_path,
        skip_list=[],
        skip_all_errors=True,
        img_dir_name=img_dir_name,
        img_public_path='',
        downloading_timeout=-1,
        deduplication=False
    )

    result = ArticleTransformer(article_path, img_downloader).run()
    print('Done.')


def main(arguments):
    """
    Entrypoint.
    """

    print(f'Markdown tool version {__version__} started...')

    article_link = arguments.article_file_path_or_url
    if is_url(article_link):
        response = download_from_url(
            article_link, timeout=arguments.downloading_timeout)
        article_path = get_filename_from_url(response)

        with open(article_path, 'wb') as article_file:
            article_file.write(response.content)
            article_file.close()
    else:
        article_path = os.path.expanduser(article_link)

    skip_list = arguments.skip_list
    skip_all = arguments.skip_all_incorrect

    print(f'File "{article_path}" will be processed...')

    if isinstance(skip_list, str):
        if skip_list.startswith('@'):
            skip_list = skip_list[1:]
            print(f'Reading skip list from a file "{skip_list}"...')
            with open(os.path.expanduser(skip_list), 'r') as fsl:
                skip_list = [s.strip() for s in fsl.readlines()]
        else:
            skip_list = [s.strip() for s in skip_list.split(',')]

    img_downloader = ImageDownloader(
        article_path=article_path,
        skip_list=skip_list,
        skip_all_errors=skip_all,
        img_dir_name=arguments.images_dirname,
        img_public_path=arguments.images_publicpath,
        downloading_timeout=arguments.downloading_timeout,
        deduplication=arguments.dedup_with_hash
    )

    result = ArticleTransformer(article_path, img_downloader).run()

    formatter = [
        f for f in FORMATTERS if f is not None and f.format == arguments.output_format]
    assert len(formatter) == 1
    formatter = formatter[0]

    article_file_name = os.path.splitext(article_path)[0]
    article_out_path = f'{article_file_name}.{formatter.format}'
    if article_path == article_out_path and not arguments.remove_source:
        article_out_path = f'{article_file_name}_{strftime("%Y%m%d_%H%M%S")}.{formatter.format}'
    print(f'Writing file into "{article_out_path}"...')

    with open(article_out_path, 'wb') as outfile:
        outfile.write(formatter.write(result))

    if arguments.remove_source and article_path != article_out_path:
        os.remove(article_path)

    print('Processing finished successfully...')


if __name__ == '__main__':
    out_format_list = [f.format for f in FORMATTERS if f is not None]

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('article_file_path_or_url', type=str,
                        help='path to the article file in the Markdown format')
    parser.add_argument('-s', '--skip-list', default=None,
                        help='skip URL\'s from the comma-separated list (or file with a leading \'@\')')
    parser.add_argument('-d', '--images-dirname', default='images',
                        help='Folder in which to download images')
    parser.add_argument('-p', '--images-publicpath', default='',
                        help='Public path to the folder of downloaded images')
    parser.add_argument('-a', '--skip-all-incorrect', default=False, action='store_true',
                        help='skip all incorrect images')
    parser.add_argument('-t', '--downloading-timeout', type=float, default=-1,
                        help='how many seconds to wait before downloading will be failed')
    parser.add_argument('-D', '--dedup-with-hash', default=False, action='store_true',
                        help='Deduplicate images, using content hash')
    parser.add_argument('-R', '--remove-source', default=False, action='store_true',
                        help='Remove or replace source file')
    parser.add_argument('-o', '--output-format', default=out_format_list[0], choices=out_format_list,
                        help='output format')
    parser.add_argument('--version', action='version',
                        version=f'%(prog)s {__version__}', help='return version number')

    args = parser.parse_args()

    main(args)




from traceback import print_tb
import open3d as o3d
from .o3dwrapper import Vector3dVector, create_mesh, load_mesh
from .utils import Timer
from .utils import BaseSocket, log
import json
import numpy as np
from os.path import join
import os

from .utils import CritRange
from .utils import get_rgb_01
from alfred.utils.base_config import load_object
import copy
from .o3d_visconfig import Config as DefaultConfig

rotate = False


def o3d_callback_rotate(vis=None):
    global rotate
    rotate = not rotate
    return False


class VisOpen3DSocket(BaseSocket):
    def __init__(self, cfg=None) -> None:
        if cfg is None:
            crt_dir = os.path.dirname(__file__)
            cfg = DefaultConfig.load(
                filename=os.path.join(crt_dir, 'default_viscfg.yml'))
        # output
        host = cfg.host
        port = cfg.port

        self.write = cfg.write
        self.out = cfg.out
        self.cfg = cfg
        if self.write:
            print('[Info] capture the screen to {}'.format(self.out))
            os.makedirs(self.out, exist_ok=True)
        # scene
        vis = o3d.visualization.VisualizerWithKeyCallback()
        vis.register_key_callback(ord('A'), o3d_callback_rotate)
        vis.create_window(window_name='Visualizer',
                          width=cfg.width, height=cfg.height)
        self.vis = vis
        # load the scene
        for sc in cfg.scene:
            key = sc['module']
            mesh_args = sc['args']
            mesh = load_object(key, mesh_args)
            self.vis.add_geometry(mesh)

        for key, val in cfg.extra.items():
            mesh = load_mesh(val["path"])
            trans = np.array(val['transform']).reshape(4, 4)
            mesh.transform(trans)
            self.vis.add_geometry(mesh)
        # create vis => update => super() init
        super().__init__(host, port, debug=cfg.debug)
        self.block = cfg.block
        if os.path.exists(cfg.body_model_template):
            body_template = o3d.io.read_triangle_mesh(cfg.body_model_template)
            self.body_template = body_template
        else:
            self.body_template = None
        self.body_model = load_object(
            cfg.body_model.module, cfg.body_model.args)
        zero_params = self.body_model.init_params(1)
        self.max_human = cfg.max_human
        self.track = cfg.track
        self.filter = cfg.filter
        self.camera_pose = cfg.camera.camera_pose
        self.init_camera(cfg.camera.camera_pose)
        self.zero_vertices = Vector3dVector(
            np.zeros((self.body_model.nVertices, 3)))

        self.vertices, self.meshes = [], []
        for i in range(self.max_human):
            self.add_human(zero_params)

        self.count = 0
        self.previous = {}
        self.critrange = CritRange(**cfg.range)
        self.new_frames = cfg.new_frames

    def add_human(self, zero_params):
        vertices = self.body_model(
            return_verts=True, return_tensor=False, **zero_params)[0]
        self.vertices.append(vertices)
        if self.body_template is None:  # create template
            mesh = create_mesh(
                vertices=vertices, faces=self.body_model.faces, colors=self.body_model.color)
        else:
            mesh = copy.deepcopy(self.body_template)
        self.meshes.append(mesh)
        self.vis.add_geometry(mesh)
        self.init_camera(self.camera_pose)

    @staticmethod
    def set_camera(cfg, camera_pose):
        theta, phi = np.deg2rad(-(cfg.camera.theta + 90)
                                ), np.deg2rad(cfg.camera.phi)
        theta = theta + np.pi
        st, ct = np.sin(theta), np.cos(theta)
        sp, cp = np.sin(phi), np.cos(phi)
        rot_x = np.array([
            [1., 0., 0.],
            [0., ct, -st],
            [0, st, ct]
        ])
        rot_z = np.array([
            [cp, -sp, 0],
            [sp, cp, 0.],
            [0., 0., 1.]
        ])
        camera_pose[:3, :3] = rot_x @ rot_z
        return camera_pose

    def init_camera(self, camera_pose):
        ctr = self.vis.get_view_control()
        init_param = ctr.convert_to_pinhole_camera_parameters()
        # init_param.intrinsic.set_intrinsics(init_param.intrinsic.width, init_param.intrinsic.height, fx, fy, cx, cy)
        init_param.extrinsic = np.array(camera_pose)
        ctr.convert_from_pinhole_camera_parameters(init_param)

    def get_camera(self):
        ctr = self.vis.get_view_control()
        init_param = ctr.convert_to_pinhole_camera_parameters()
        return np.array(init_param.extrinsic)

    def filter_human(self, datas):
        datas_new = []
        for data in datas:
            kpts3d = np.array(data['keypoints3d'])
            data['keypoints3d'] = kpts3d
            pid = data['id']
            if pid not in self.previous.keys():
                if not self.critrange(kpts3d):
                    continue
                self.previous[pid] = 0
            self.previous[pid] += 1
            if self.previous[pid] > self.new_frames:
                datas_new.append(data)
        return datas_new

    def main(self, datas):
        if self.debug:
            log('[Info] Load data {}'.format(self.count))
        if isinstance(datas, str):
            datas = json.loads(datas)
        print(datas)
        for data in datas:
            for key in data.keys():
                if key == 'id':
                    continue
                data[key] = np.array(data[key])
            if 'keypoints3d' not in data.keys() and self.filter:
                data['keypoints3d'] = self.body_model(
                    return_verts=False, return_tensor=False, **data)[0]
        if self.filter:
            datas = self.filter_human(datas)
        with Timer('forward'):
            params = []
            for i, data in enumerate(datas):
                if i >= len(self.meshes):
                    print('[Error] the number of human exceeds!')
                    self.add_human(data)
                if 'vertices' in data.keys():
                    vertices = data['vertices']
                    self.vertices[i] = Vector3dVector(vertices)
                else:
                    params.append(data)
            if len(params) > 0:
                params = self.body_model.merge_params(
                    params, share_shape=False)
                vertices = self.body_model(
                    return_verts=True, return_tensor=False, **params)
                for i in range(vertices.shape[0]):
                    self.vertices[i] = Vector3dVector(vertices[i])
            for i in range(len(datas), len(self.meshes)):
                self.vertices[i] = self.zero_vertices
        # Open3D will lock the thread here
        with Timer('set vertices'):
            for i in range(len(self.vertices)):
                self.meshes[i].vertices = self.vertices[i]
                if i < len(datas) and self.track:
                    col = get_rgb_01(datas[i]['id'])
                    self.meshes[i].paint_uniform_color(col)
        print('verts done.')

    def o3dcallback(self):
        if rotate:
            self.cfg.camera.phi += np.pi/10
            camera_pose = self.set_camera(self.cfg, self.get_camera())
            self.init_camera(camera_pose)

    def update(self):
        if self.disconnect and not self.block:
            self.previous.clear()
        if not self.queue.empty():
            if self.debug:
                log('Update' + str(self.queue.qsize()))
            datas = self.queue.get()
            if not self.block:
                while self.queue.qsize() > 0:
                    datas = self.queue.get()
            self.main(datas)
            with Timer('update geometry'):
                for mesh in self.meshes:
                    mesh.compute_triangle_normals()
                    self.vis.update_geometry(mesh)
                self.o3dcallback()
                self.vis.poll_events()
                self.vis.update_renderer()
            if self.write:
                outname = join(self.out, '{:06d}.jpg'.format(self.count))
                with Timer('capture'):
                    self.vis.capture_screen_image(outname)
            self.count += 1
        else:
            with Timer('update renderer', True):
                self.o3dcallback()
                self.vis.poll_events()
                self.vis.update_renderer()



# -*- coding: utf-8 -*-
# brought from https://github.com/mkocabas/VIBE/blob/master/lib/utils/renderer.py
import sys
import os
import json
from typing import Tuple
from pytorch3d.renderer.cameras import PerspectiveCameras
import torch
from torch import nn
import pickle

# Data structures and functions for rendering
from pytorch3d.structures import Meshes, join_meshes_as_scene
from pytorch3d.renderer import (
    look_at_view_transform,
    FoVPerspectiveCameras,
    FoVOrthographicCameras,
    OpenGLPerspectiveCameras,
    PointLights,
    DirectionalLights,
    Materials,
    RasterizationSettings,
    MeshRenderer,
    MeshRasterizer,
    SoftPhongShader,
    TexturesUV,
    TexturesVertex,
)
# from pytorch3d.io.mesh
from pytorch3d.io import IO
import numpy as np


"""
source code adopt from ROMP
edited by alfred-py
"""


colors = {
    "pink": [0.7, 0.7, 0.9],
    "neutral": [0.9, 0.9, 0.8],
    "capsule": [0.7, 0.75, 0.5],
    "yellow": [0.5, 0.7, 0.75],
}


def get_projection_matrix_for_weak_perspective_camera(s_x, s_y, t_x, t_y):
    P = torch.eye(4)
    P[0, 0] = s_x
    P[1, 1] = s_y
    P[0, 3] = t_x * s_x
    P[1, 3] = -t_y * s_y
    P[2, 2] = -1
    print(P)
    return P[:-1, :-1], P[:, -1][:-1]


class Renderer(nn.Module):
    def __init__(
        self,
        faces,
        resolution=(512, 512),
        perps=True,
        R=None,
        T=None,
        use_gpu=False,
    ):
        super(Renderer, self).__init__()
        self.name = 'pytorch3d'
        self.perps = perps
        if use_gpu:
            self.device = torch.device('cuda')
            print("visualize in gpu mode")
        else:
            self.device = torch.device("cpu")
            print("visualize in cpu mode")

        if isinstance(faces, np.ndarray):
            faces = torch.from_numpy(faces.astype(np.float))
            faces.to(self.device)
        self.faces = faces.unsqueeze(0).to(
            self.device)  # add a BatchSize dim
        self.default_color = torch.as_tensor(colors["neutral"]).unsqueeze(0)
        print('default color: ', self.default_color.shape)
        self.save_io = IO()

        if R is None:
            self.default_R = torch.Tensor(
                [[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]])
        if T is None:
            self.default_T = torch.Tensor([[0.0, 0.0, 0.0]])
        self.fov = 60

        if self.perps:
            # self.cameras = FoVPerspectiveCameras(zfar=1000,
            #     R=self.default_R, T=self.default_T, fov=60, device=self.device
            # )
            self.cameras = PerspectiveCameras(focal_length=5000,
                R=self.default_R, T=self.default_T, device=self.device
            )
            self.lights = PointLights(
                ambient_color=((0.56, 0.56, 0.56),),
                location=torch.Tensor([[0.0, 0.0, 0.0]]),
                device=self.device,
            )
        else:
            self.cameras = FoVOrthographicCameras(
                R=self.default_R,
                T=self.default_T,
                znear=0.0,
                zfar=100.0,
                max_y=1.0,
                min_y=-1.0,
                max_x=1.0,
                min_x=-1.0,
                device=self.device,
            )
            self.lights = DirectionalLights(
                direction=torch.Tensor([[0.0, 1.0, 0.0]]), device=self.device
            )

        # Define the settings for rasterization and shading. Here we set the output image to be of size
        # 512x512. As we are rendering images for visualization purposes only we will set faces_per_pixel=1
        # and blur_radius=0.0.
        raster_settings = RasterizationSettings(
            image_size=(resolution[0], resolution[1]), blur_radius=0.0, faces_per_pixel=1
        )
        print('resolution: ', resolution)

        # Create a Phong renderer by composing a rasterizer and a shader. The textured Phong shader will
        # interpolate the texture uv coordinates for each vertex, sample from a texture image and
        # apply the Phong lighting model
        self.renderer = MeshRenderer(
            rasterizer=MeshRasterizer(
                cameras=self.cameras, raster_settings=raster_settings
            ),
            shader=SoftPhongShader(
                device=self.device, cameras=self.cameras, lights=self.lights
            ),
        )

    # def render(self, img, verts, cam, angle=None, axis=None, mesh_filename=None, color=[1.0, 1.0, 0.9], rotate=False):
    def render(
        self, img, verts, faces=None, color=None, 
        camera_t=torch.zeros([1, 3, 3], dtype=torch.float32),
        camera_rot=torch.zeros([1, 3, 3], dtype=torch.float32),
        focal_length=1000,
        merge_meshes=True, cam=None, mesh_filename=None, **kwargs
    ):
        assert len(
            verts.shape) >= 2, f"The input verts of visualizer is bounded to be 3-dims (Nx6890 x3) tensor, but got: {verts.shape}"
        if isinstance(verts, np.ndarray):
            verts = torch.from_numpy(verts)
        if len(verts.shape) == 2:
            verts = verts.unsqueeze(0)
        if faces is None:
            faces = self.faces
        
        BS = verts.shape[0]
        if verts.shape[0] != faces.shape[0]:
            faces = faces.repeat(BS, 1, 1).to(self.device)

        verts = verts.to(self.device)

        if isinstance(color, np.ndarray) or isinstance(color, tuple) or isinstance(color, list):
            # color = torch.from_numpy(np.array(color)).to(self.device).unsqueeze(1)
            color = torch.from_numpy(np.array(color)).to(
                self.device).unsqueeze(0)
        elif color is None:
            color = self.default_color

        verts = verts.to(self.device)
        verts = verts.float()
        verts_rgb = torch.ones_like(verts)
        verts_rgb[:, :] = color
        textures = TexturesVertex(verts_features=verts_rgb)
        verts[:, :, :2] *= -1
        meshes = Meshes(verts, faces, textures)
        if merge_meshes:
            meshes = join_meshes_as_scene(meshes)
        if mesh_filename is not None:
            self.save_io.save_mesh(meshes, mesh_filename)
        if camera_t is not None:
            # cam = cam.float()
            # print(cam)
            cam = torch.as_tensor(camera_t).to(self.device)
            if self.perps:
                T = cam
                T = T.unsqueeze(0).repeat(BS, 1)
                print(T)
                # new_cam = FoVPerspectiveCameras(zfar=1000, znear=0.05,
                #     R=self.default_R, T=T, fov=self.fov, device=self.device)
                new_cam = PerspectiveCameras(focal_length=focal_length,
                    R=self.default_R, T=T, device=self.device)
            else:
                R, T, xyz_ranges = cam
                new_cam = FoVOrthographicCameras(
                    R=R, T=self.default_T, **xyz_ranges, device=self.device
                )
            images = self.renderer(meshes, cameras=new_cam)
        else:
            images = self.renderer(meshes)
        print(images)
        images[..., :-1] *= 255
        images = images[..., :-1].cpu().numpy()
        print(images)
        if images.shape[-1] > 3:
            images = images[..., :-1]
        return images


def get_renderer(test=False, **kwargs):
    renderer = Renderer(**kwargs)
    if test:
        import cv2

        dist = 1 / np.tan(np.radians(args().FOV / 2.0))
        print("dist:", dist)
        model = pickle.load(
            open(
                os.path.join(args().smpl_model_path, "smpl",
                             "SMPL_NEUTRAL.pkl"), "rb"
            ),
            encoding="latin1",
        )
        np_v_template = (
            torch.from_numpy(
                np.array(model["v_template"])).cuda().float()[None]
        )
        face = torch.from_numpy(model["f"].astype(np.int32)).cuda()[None]
        np_v_template = np_v_template.repeat(2, 1, 1)
        np_v_template[1] += 0.3
        np_v_template[:, :, 2] += dist
        face = face.repeat(2, 1, 1)
        result = renderer(np_v_template, face).cpu().numpy()
        for ri in range(len(result)):
            cv2.imwrite(
                "test{}.png".format(
                    ri), (result[ri, :, :, :3] * 255).astype(np.uint8)
            )
    return renderer


if __name__ == "__main__":
    get_renderer(test=True, perps=True)



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
Utils using in MANA universe

such as print welcome message
"""
from colorama import Fore, Back, Style


welcome_msg = '''
    __  ______    _   _____    ___    ____
   /  |/  /   |  / | / /   |  /   |  /  _/
  / /|_/ / /| | /  |/ / /| | / /| |  / /  
 / /  / / ___ |/ /|  / ___ |/ ___ |_/ /   
/_/  /_/_/  |_/_/ |_/_/  |_/_/  |_/___/    http://manaai.cn
'''

def welcome(ori_git_url):
    print(Fore.YELLOW + Style.BRIGHT + 'Welcome to MANA AI platform!' + Style.RESET_ALL)
    print(Fore.BLUE + Style.BRIGHT + welcome_msg + Style.RESET_ALL)
    print(Style.BRIGHT + "once you saw this msg, indicates you were back supported by our team!" + Style.RESET_ALL)
    print('the latest updates of our codes always at: {} or {}'.format(ori_git_url, 'http://manaai.cn'))
    print('NOTE: Our codes distributed from anywhere else were not supported!')



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
this file contains a tool kit for computer vision kit
we can using this powerful kit to display detection or
segmentation skillfully
"""
import numpy as np
import colorsys
import cv2
import time
import os
import tensorflow as tf
from PIL import Image
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.patheffects as patheffects


def draw_box_without_score(img, boxes, classes=None, is_show=False):
    """
    Draw boxes on image, the box mostly are annotations, not the model predict box
    """
    assert isinstance(boxes,
                      np.ndarray), 'boxes must nump array, with shape of (None, 5)\nevery element contains (x1,y1,x2,y2, label)'
    if classes:
        pass
    else:
        height = img.shape[0]
        width = img.shape[1]

        font = cv2.QT_FONT_NORMAL
        font_scale = 0.4
        font_thickness = 1
        line_thickness = 1

        all_cls = []
        for i in range(boxes.shape[0]):
            cls = boxes[i, -1]
            all_cls.append(cls)
            all_cls = set(all_cls)
            unique_color = _create_unique_color_uchar(all_cls.index(cls))

            y1 = int(boxes[i, 2])
            x1 = int(boxes[i, 3])
            y2 = int(boxes[i, 4])
            x2 = int(boxes[i, 5])

            cv2.rectangle(img, (x1, y1), (x2, y2), unique_color, line_thickness)

            text_label = '{}'.format(cls)
            (ret_val, base_line) = cv2.getTextSize(text_label, font, font_scale, font_thickness)
            text_org = (x1, y1 - 0)

            cv2.rectangle(img, (text_org[0] - 5, text_org[1] + base_line + 2),
                          (text_org[0] + ret_val[0] + 5, text_org[1] - ret_val[1] - 2), unique_color, line_thickness)
            # this rectangle for fill text rect
            cv2.rectangle(img, (text_org[0] - 5, text_org[1] + base_line + 2),
                          (text_org[0] + ret_val[0] + 4, text_org[1] - ret_val[1] - 2),
                          unique_color, -1)
            cv2.putText(img, text_label, text_org, font, font_scale, (255, 255, 255), font_thickness)
        if is_show:
            cv2.imshow('image', img)
            cv2.waitKey(0)
        return img


def visualize_det_cv2(img, detections, classes=None, thresh=0.6, is_show=False, background_id=-1):
    """
    visualize detection on image using cv2, this is the standard way to visualize detections
    :param img:
    :param detections: ssd detections, numpy.array([[id, score, x1, y1, x2, y2]...])
            each row is one object
    :param classes:
    :param thresh:
    :param is_show:
    :param background_id: -1
    :return:
    """
    assert classes, 'from visualize_det_cv2, classes must be provided, each class in a list with' \
                    'certain order.'
    assert isinstance(img, np.ndarray), 'from visualize_det_cv2, img must be a numpy array object.'

    height = img.shape[0]
    width = img.shape[1]

    font = cv2.QT_FONT_NORMAL
    font_scale = 0.4
    font_thickness = 1
    line_thickness = 1

    for i in range(detections.shape[0]):
        cls_id = int(detections[i, 0])
        if cls_id != background_id:
            score = detections[i, 1]
            if score > thresh:
                unique_color = _create_unique_color_uchar(cls_id)

                # if detection coordinates normalized, then do this step, otherwise not
                # x1 = int(detections[i, 2] * width)
                # y1 = int(detections[i, 3] * height)
                # x2 = int(detections[i, 4] * width)
                # y2 = int(detections[i, 5] * height)

                y1 = int(detections[i, 2])
                x1 = int(detections[i, 3])
                y2 = int(detections[i, 4])
                x2 = int(detections[i, 5])

                cv2.rectangle(img, (x1, y1), (x2, y2), unique_color, line_thickness)

                text_label = '{} {:.2f}'.format(classes[cls_id], score)
                (ret_val, base_line) = cv2.getTextSize(text_label, font, font_scale, font_thickness)
                text_org = (x1, y1 - 0)

                cv2.rectangle(img, (text_org[0] - 5, text_org[1] + base_line + 2),
                              (text_org[0] + ret_val[0] + 5, text_org[1] - ret_val[1] - 2), unique_color,
                              line_thickness)
                # this rectangle for fill text rect
                cv2.rectangle(img, (text_org[0] - 5, text_org[1] + base_line + 2),
                              (text_org[0] + ret_val[0] + 4, text_org[1] - ret_val[1] - 2),
                              unique_color, -1)
                cv2.putText(img, text_label, text_org, font, font_scale, (255, 255, 255), font_thickness)
    if is_show:
        cv2.imshow('image', img)
        cv2.waitKey(0)
    return img


def visualize_det_mask_cv2(img, detections, masks, classes=None, is_show=False, background_id=-1, is_video=False):
    """
    this method using for display detections and masks on image
    :param img:
    :param detections: numpy.array([[id, score, x1, y1, x2, y2]...])
            each row is one object. contains id and score in the first 2 rows
    :param masks: numpy.array([[mask_width, mask_height], ...], every element is an
    one chanel mask of on object
    :param classes: classes names in a list with certain order
    :param is_show: to show if it is video
    :param background_id
    :param is_video
    :return:
    """
    assert isinstance(img, np.ndarray) and isinstance(detections, np.ndarray) and isinstance(masks, np.ndarray), \
        'images and detections and masks must be numpy array'
    assert detections.shape[0] == masks.shape[-1], 'detections nums and masks nums are not equal'
    assert is_show != is_video, 'you can not set is_show and is_video at the same time.'
    # draw detections first
    img = visualize_det_cv2(img, detections, classes=classes, is_show=False)

    masked_image = img
    print('masked image shape: ', masked_image.shape)
    num_instances = detections.shape[0]
    for i in range(num_instances):
        cls_id = int(detections[i, 0])
        if cls_id != background_id:
            unique_color = _create_unique_color_uchar(cls_id)
            mask = masks[:, :, i]
            masked_image = _apply_mask2(masked_image, mask, unique_color)
    # masked_image = masked_image.astype(int)
    if is_video:
        cv2.imshow('image', masked_image)
        cv2.waitKey(1)
    elif is_show:
        cv2.imshow('image', masked_image)
        cv2.waitKey(0)
    return masked_image


def draw_masks(img, masks, cls_color_list, is_show=False, background_id=-1, is_video=False, convert_bgr=False):
    """
    draw masks pure on an image, the mask format is something like this:
    [[[1], [1], [1], .., [2]],
     [[1], [1], [1], .., [2]],
     [[1], [1], [1], .., [2]]]
    every pixel in image is a class

    the color list better using RGBA channel
    cls_color_list = [(223,  224, 225, 0.4), (12, 23, 23, 0.4), ...] a list of colors

    Note: suppose the img in BGR format, you should convert to RGB once img returned
    :param img:
    :param masks:
    :param cls_color_list:
    :param is_show:
    :param background_id:
    :param is_video:
    :return:
    """
    n, h, w, c = masks.shape

    mask_flatten = masks[0].flatten()
    mask_color = np.array(list(map(lambda i: cls_color_list[i], mask_flatten)))
    # reshape to normal image shape,
    mask_color = np.reshape(mask_color, (h, w, 3)).astype('float32')

    # add this mask on img
    # img = cv2.add(img, mask_color)
    if convert_bgr:
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    img = cv2.addWeighted(img, 0.6, mask_color, 0.4, 0)
    if is_show:
        cv2.imshow('img', img)
        cv2.imwrite('test_res.jpg', img)
        cv2.waitKey(0)
    return img


def _apply_mask2(image, mask, color, alpha=0.5):
    for c in range(3):
        image[:, :, c] = np.where(mask == 1, image[:, :, c] * (1 - alpha) + alpha * color[c],
                                  image[:, :, c])
    return image


def create_unique_color_float(tag, hue_step=0.41, alpha=0.7):
    h, v = (tag * hue_step) % 1, 1. - (int(tag * hue_step) % 4) / 5.
    r, g, b = colorsys.hsv_to_rgb(h, 1., v)
    return r, g, b, alpha


def create_unique_color_uchar(tag, hue_step=0.41, alpha=0.7):
    r, g, b, a = create_unique_color_float(tag, hue_step, alpha)
    return int(255 * r), int(255 * g), int(255 * b), int(255 * a)


# ----------------------- 3D drawing functionality ----------------------




#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
convert labelone to voc format

<annotation>
  <folder>VOC2007</folder>
  <filename>000002.jpg</filename>    //文件名  
  <size>                            //图像尺寸（长宽以及通道数
    <width>335</width>
    <height>500</height>
    <depth>3</depth>
  </size>
  <object>        //检测到的物体
    <name>cat</name>    //物体类别
    <pose>Unspecified</pose>    //拍摄角度
    <truncated>0</truncated>    //是否被截断（0表示完整
    <difficult>0</difficult>    //目标是否难以识别（0表示容易识别）
    <bndbox>                    //bounding-box（包含左下角和右上角xy坐标）
      <xmin>139</xmin>
      <ymin>200</ymin>
      <xmax>207</xmax>
      <ymax>301</ymax>
    </bndbox>
  </object>
</annotation>


"""
import os
import json
import glob
import sys
from lxml.etree import Element, SubElement, tostring, ElementTree, tostring
from lxml import etree
from PIL import Image


def convert_one(a):
    os.makedirs(os.path.dirname(a)+'_voc', exist_ok=True)
    d = json.load(open(a))
    print(d)

    target_path = os.path.join(os.path.dirname(a)+'_voc', os.path.basename(a).split('.')[0]+'.xml')
    img_path = os.path.join('images', d['imagePath'])
    # convert xml 
    if os.path.exists(img_path):
        im = Image.open(img_path)
        width = im.size[0]
        height = im.size[1]
        node_root = Element('annotation')
        node_folder = SubElement(node_root, 'folder')
        node_folder.text = 'images'
        node_filename = SubElement(node_root, 'filename')
        node_filename.text = d['imagePath']
        node_size = SubElement(node_root, 'size')
        node_width = SubElement(node_size, 'width')
        node_width.text = str(width)
        node_height = SubElement(node_size, 'height')
        node_height.text = str(height)
        node_depth = SubElement(node_size, 'depth')
        node_depth.text = '3'
        
        for item in d['shapes']:
            node_object = SubElement(node_root, 'object')
            node_name = SubElement(node_object, 'name')
            node_name.text = item['label']
            node_difficult = SubElement(node_object, 'difficult')
            node_difficult.text = '0'
            node_bndbox = SubElement(node_object, 'bndbox')
            node_xmin = SubElement(node_bndbox, 'xmin')
            node_xmin.text = str(item['points'][1][0])
            node_ymin = SubElement(node_bndbox, 'ymin')
            node_ymin.text = str(item['points'][1][1])
            node_xmax = SubElement(node_bndbox, 'xmax')
            node_xmax.text = str(item['points'][3][0])
            node_ymax = SubElement(node_bndbox, 'ymax')
            node_ymax.text = str(item['points'][3][1])
        f = open(target_path, 'wb')
        f.write(etree.tostring(node_root, pretty_print=True))
        f.close()
    else:
        print('xxx {} annotations according image: {} not exist.'.format(a, img_path))



def run():
    all_json_files = glob.glob(os.path.join(sys.argv[1], '*.json'))
    print(len(all_json_files))
    for i in all_json_files:
        convert_one(i)
    print('done!')



if __name__ == "__main__":
    run()